# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import pyserum.enums
import typing

from datetime import datetime
from decimal import Decimal
from pyserum._layouts.instructions import (
    INSTRUCTIONS_LAYOUT as PYSERUM_INSTRUCTIONS_LAYOUT,
    InstructionType as PySerumInstructionType,
)
from pyserum.enums import OrderType as PySerumOrderType, Side as PySerumSide
from pyserum.instructions import (
    settle_funds as pyserum_settle_funds,
    SettleFundsParams as PySerumSettleFundsParams,
)
from pyserum.market.market import Market as PySerumMarket
from pyserum.open_orders_account import (
    make_create_account_instruction as pyserum_make_create_account_instruction,
)
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.system_program import CreateAccountParams, create_account
from solana.transaction import AccountMeta, TransactionInstruction
from spl.token.constants import (
    ACCOUNT_LEN,
    TOKEN_PROGRAM_ID,
)
from spl.token.instructions import (
    CloseAccountParams,
    InitializeAccountParams,
    TransferParams,
    close_account,
    create_associated_token_account,
    initialize_account,
    transfer,
)

from .combinableinstructions import CombinableInstructions
from .constants import I64_MAX, SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .layouts import layouts
from .orders import Order, OrderType, Side

from .perpmarketdetails import PerpMarketDetails

# from .spotmarket import SpotMarket
from .tokens import Token
from .tokenaccount import TokenAccount
from .tokenbank import TokenBank, NodeBank, RootBank
from .wallet import Wallet

# 🥭 Interfaces
#
# To avoid circular dependencies, here we specify the bare interfaces we need for some
# of the objects. For instance, where a function takes an IAccount, you can (and probably
# should) safely pass in an IAccount.
#


class IAccount(typing.Protocol):
    group_address: PublicKey

    @property
    def address(self) -> PublicKey:
        raise NotImplementedError(
            "IAccount.address is not implemented on the Protocol."
        )

    @property
    def spot_open_orders_by_index(self) -> typing.Sequence[typing.Optional[PublicKey]]:
        raise NotImplementedError(
            "IAccount.spot_open_orders_by_index is not implemented on the Protocol."
        )

    @property
    def spot_open_orders(self) -> typing.Sequence[PublicKey]:
        raise NotImplementedError(
            "IAccount.spot_open_orders is not implemented on the Protocol."
        )

    def update_spot_open_orders_for_market(
        self, spot_market_index: int, spot_open_orders: PublicKey
    ) -> None:
        raise NotImplementedError(
            "IAccount.update_spot_open_orders_for_market() is not implemented on the Protocol."
        )


class IGroupSlot(typing.Protocol):
    index: int


class IGroup(typing.Protocol):
    cache: PublicKey
    signer_key: PublicKey
    fees_vault: PublicKey
    shared_quote: TokenBank

    @property
    def address(self) -> PublicKey:
        raise NotImplementedError("IGroup.address is not implemented on the Protocol.")

    @property
    def tokens_by_index(self) -> typing.Sequence[typing.Optional[TokenBank]]:
        raise NotImplementedError(
            "IGroup.tokens_by_index is not implemented on the Protocol."
        )

    @property
    def base_tokens_by_index(self) -> typing.Sequence[typing.Optional[TokenBank]]:
        raise NotImplementedError(
            "IGroup.base_tokens_by_index is not implemented on the Protocol."
        )

    def slot_by_spot_market_address(self, spot_market_address: PublicKey) -> IGroupSlot:
        raise NotImplementedError(
            "IGroup.slot_by_spot_market_address() is not implemented on the Protocol."
        )


class IPerpMarket(typing.Protocol):
    underlying_perp_market: PerpMarketDetails

    @property
    def address(self) -> PublicKey:
        raise NotImplementedError(
            "IPerpMarket.address is not implemented on the Protocol."
        )


class ISpotMarket(typing.Protocol):
    underlying_serum_market: PySerumMarket

    @property
    def address(self) -> PublicKey:
        raise NotImplementedError(
            "ISpotMarket.address is not implemented on the Protocol."
        )

    @property
    def bids_address(self) -> PublicKey:
        raise NotImplementedError(
            "ISpotMarket.bids_address is not implemented on the Protocol."
        )

    @property
    def asks_address(self) -> PublicKey:
        raise NotImplementedError(
            "ISpotMarket.asks_address is not implemented on the Protocol."
        )

    @property
    def event_queue_address(self) -> PublicKey:
        raise NotImplementedError(
            "ISpotMarket.event_queue_address is not implemented on the Protocol."
        )


# 🥭 Instructions
#
# This file contains the low-level instruction functions that build the raw instructions
# to send to Solana.
#
# One important distinction between these functions and the more common `create instruction functions` in
# Solana is that these functions *all return a combinable of instructions and signers*.
#
# It's likely that some operations will require actions split across multiple instructions because of
# instruction size limitiations, so all our functions are prepared for this without having to change
# the function signature in future.
#

# #
# # 🥭 SOLANA instruction builders
# #

# # 🥭 build_solana_create_account_instructions function
#
# Creates and initializes an SPL token account. Can add additional lamports too but that's usually not
# necesary.
#
def build_solana_create_account_instructions(
    context: Context,
    wallet: Wallet,
    entropy_program_address: PublicKey,
    size: int,
    lamports: int = 0,
) -> CombinableInstructions:
    minimum_balance = context.client.get_minimum_balance_for_rent_exemption(size)
    account = Keypair()
    create_instruction = create_account(
        CreateAccountParams(
            wallet.address,
            account.public_key,
            lamports + minimum_balance,
            size,
            entropy_program_address,
        )
    )
    return CombinableInstructions(signers=[account], instructions=[create_instruction])


# #
# # 🥭 SPL instruction builders
# #


# # 🥭 build_spl_create_account_instructions function
#
# Creates and initializes an SPL token account. Can add additional lamports too but that's usually not
# necesary.
#
# Prefer `build_spl_create_account_instructions()` over this function. This function should be
# reserved for cases where you specifically don't want the associated token account.
#
def build_spl_create_account_instructions(
    context: Context, wallet: Wallet, token: Token, lamports: int = 0
) -> CombinableInstructions:
    create_account_instructions = build_solana_create_account_instructions(
        context, wallet, TOKEN_PROGRAM_ID, ACCOUNT_LEN, lamports
    )
    initialize_instruction = initialize_account(
        InitializeAccountParams(
            TOKEN_PROGRAM_ID,
            create_account_instructions.signers[0].public_key,
            token.mint,
            wallet.address,
        )
    )
    return create_account_instructions + CombinableInstructions(
        signers=[], instructions=[initialize_instruction]
    )


# # 🥭 build_spl_create_associated_account_instructions function
#
# Creates and initializes an 'associated' SPL token account. This is the usual way of creating a
# token account now. `build_spl_create_account_instructions()` should be reserved for cases where
# you specifically don't want the associated token account.
#
def build_spl_create_associated_account_instructions(
    context: Context, wallet: Wallet, owner: PublicKey, token: Token
) -> CombinableInstructions:
    create_account_instructions = create_associated_token_account(
        wallet.address, owner, token.mint
    )
    return CombinableInstructions(
        signers=[], instructions=[create_account_instructions]
    )


# # 🥭 build_spl_transfer_tokens_instructions function
#
# Creates an instruction to transfer SPL tokens from one account to another.
#
def build_spl_transfer_tokens_instructions(
    context: Context,
    wallet: Wallet,
    token: Token,
    source: PublicKey,
    destination: PublicKey,
    quantity: Decimal,
) -> CombinableInstructions:
    amount = int(token.shift_to_native(quantity))
    instructions = [
        transfer(
            TransferParams(
                TOKEN_PROGRAM_ID, source, destination, wallet.address, amount, []
            )
        )
    ]
    return CombinableInstructions(signers=[], instructions=instructions)


# # 🥭 build_spl_close_account_instructions function
#
# Creates an instructio to close an SPL token account and transfers any remaining lamports to the wallet.
#
def build_spl_close_account_instructions(
    context: Context, wallet: Wallet, address: PublicKey
) -> CombinableInstructions:
    return CombinableInstructions(
        signers=[],
        instructions=[
            close_account(
                CloseAccountParams(
                    TOKEN_PROGRAM_ID, address, wallet.address, wallet.address
                )
            )
        ],
    )


# # 🥭 build_spl_faucet_airdrop_instructions function
#
# Creates an airdrop instruction for compatible faucets (those based on https://github.com/paul-schaaf/spl-token-faucet)
#
def build_spl_faucet_airdrop_instructions(
    token_mint: PublicKey, destination: PublicKey, faucet: PublicKey, quantity: Decimal
) -> CombinableInstructions:
    faucet_program_address: PublicKey = PublicKey(
        "4bXpkKSV8swHSnwqtzuboGPaPDeEgAn4Vt8GfarV5rZt"
    )
    authority_and_nonce: typing.Tuple[PublicKey, int] = PublicKey.find_program_address(
        [b"faucet"], faucet_program_address
    )
    authority: PublicKey = authority_and_nonce[0]

    # Mints and airdrops tokens from a faucet.
    #
    # SPL instruction is at:
    #   https://github.com/paul-schaaf/spl-token-faucet/blob/main/src/program/src/instruction.rs
    #
    # ///
    # /// Mints Tokens
    # ///
    # /// 0. `[]` The mint authority - Program Derived Address
    # /// 1. `[writable]` Token Mint IAccount
    # /// 2. `[writable]` Destination IAccount
    # /// 3. `[]` The SPL Token Program
    # /// 4. `[]` The Faucet IAccount
    # /// 5. `[optional/signer]` Admin IAccount
    faucet_airdrop_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=authority),
            AccountMeta(is_signer=False, is_writable=True, pubkey=token_mint),
            AccountMeta(is_signer=False, is_writable=True, pubkey=destination),
            AccountMeta(is_signer=False, is_writable=False, pubkey=TOKEN_PROGRAM_ID),
            AccountMeta(is_signer=False, is_writable=False, pubkey=faucet),
        ],
        program_id=faucet_program_address,
        data=layouts.FAUCET_AIRDROP.build({"quantity": quantity}),
    )
    return CombinableInstructions(signers=[], instructions=[faucet_airdrop_instruction])


# #
# # 🥭 Serum instruction builders
# #


# # 🥭 build_serum_create_openorders_instructions function
#
# Creates a Serum openorders-creating instruction.
#
def build_serum_create_openorders_instructions(
    context: Context, wallet: Wallet, market: PySerumMarket
) -> CombinableInstructions:
    new_open_orders_account = Keypair()
    minimum_balance = context.client.get_minimum_balance_for_rent_exemption(
        layouts.OPEN_ORDERS.sizeof()
    )
    instruction = pyserum_make_create_account_instruction(
        owner_address=wallet.address,
        new_account_address=new_open_orders_account.public_key,
        lamports=minimum_balance,
        program_id=market.state.program_id(),
    )

    return CombinableInstructions(
        signers=[new_open_orders_account], instructions=[instruction]
    )


# # 🥭 build_serum_place_order_instructions function
#
# Creates a Serum order-placing instruction using V3 of the NewOrder instruction.
#
def build_serum_place_order_instructions(
    context: Context,
    wallet: Wallet,
    market: PySerumMarket,
    source: PublicKey,
    open_orders_address: PublicKey,
    order_type: OrderType,
    side: Side,
    price: Decimal,
    quantity: Decimal,
    client_id: int,
    fee_discount_address: PublicKey,
) -> CombinableInstructions:
    serum_order_type: PySerumOrderType = (
        PySerumOrderType.POST_ONLY
        if order_type == OrderType.POST_ONLY
        else PySerumOrderType.IOC
        if order_type == OrderType.IOC
        else PySerumOrderType.LIMIT
    )
    serum_side: PySerumSide = PySerumSide.SELL if side == Side.SELL else PySerumSide.BUY

    instruction = market.make_place_order_instruction(
        source,
        wallet.keypair,
        serum_order_type,
        serum_side,
        float(price),
        float(quantity),
        client_id,
        open_orders_address,
        fee_discount_address,
    )

    return CombinableInstructions(signers=[], instructions=[instruction])


# # 🥭 build_serum_consume_events_instructions function
#
# Creates an event-consuming 'crank' instruction.
#
def build_serum_consume_events_instructions(
    context: Context,
    market_address: PublicKey,
    event_queue_address: PublicKey,
    open_orders_addresses: typing.Sequence[PublicKey],
    limit: int = 32,
) -> CombinableInstructions:
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(pubkey=pubkey, is_signer=False, is_writable=True)
            for pubkey in [*open_orders_addresses, market_address, event_queue_address]
        ],
        program_id=context.serum_program_address,
        data=PYSERUM_INSTRUCTIONS_LAYOUT.build(
            dict(
                instruction_type=PySerumInstructionType.CONSUME_EVENTS,
                args=dict(limit=limit),
            )
        ),
    )

    # The interface accepts (and currently requires) two accounts at the end, but
    # it doesn't actually use them.
    random_account = Keypair().public_key
    instruction.keys.append(
        AccountMeta(random_account, is_signer=False, is_writable=True)
    )
    instruction.keys.append(
        AccountMeta(random_account, is_signer=False, is_writable=True)
    )
    return CombinableInstructions(signers=[], instructions=[instruction])


# # 🥭 build_serum_settle_instructions function
#
# Creates a 'settle' instruction.
#
def build_serum_settle_instructions(
    context: Context,
    wallet: Wallet,
    market: PySerumMarket,
    open_orders_address: PublicKey,
    base_token_account_address: PublicKey,
    quote_token_account_address: PublicKey,
) -> CombinableInstructions:
    vault_signer = PublicKey.create_program_address(
        [
            bytes(market.state.public_key()),
            market.state.vault_signer_nonce().to_bytes(8, byteorder="little"),
        ],
        market.state.program_id(),
    )
    instruction = pyserum_settle_funds(
        PySerumSettleFundsParams(
            market=market.state.public_key(),
            open_orders=open_orders_address,
            owner=wallet.address,
            base_vault=market.state.base_vault(),
            quote_vault=market.state.quote_vault(),
            base_wallet=base_token_account_address,
            quote_wallet=quote_token_account_address,
            vault_signer=vault_signer,
            program_id=market.state.program_id(),
        )
    )

    return CombinableInstructions(signers=[], instructions=[instruction])


# #
# # 🥭 Spot instruction builders
# #


# # 🥭 build_spot_settle_instructions function
#
# Creates a 'settle' instruction for spot markets.
#
# /// Settle all funds from serum dex open orders
# ///
# /// Accounts expected by this instruction (18):
# ///
# /// 0. `[]` entropy_group_ai - EntropyGroup that this entropy account is for
# /// 1. `[]` entropy_cache_ai - EntropyCache for this EntropyGroup
# /// 2. `[signer]` owner_ai - EntropyAccount owner
# /// 3. `[writable]` entropy_account_ai - EntropyAccount
# /// 4. `[]` dex_prog_ai - program id of serum dex
# /// 5.  `[writable]` spot_market_ai - dex MarketState account
# /// 6.  `[writable]` open_orders_ai - open orders for this market for this EntropyAccount
# /// 7. `[]` signer_ai - EntropyGroup signer key
# /// 8. `[writable]` dex_base_ai - base vault for dex MarketState
# /// 9. `[writable]` dex_quote_ai - quote vault for dex MarketState
# /// 10. `[]` base_root_bank_ai - EntropyGroup base vault acc
# /// 11. `[writable]` base_node_bank_ai - EntropyGroup quote vault acc
# /// 12. `[]` quote_root_bank_ai - EntropyGroup quote vault acc
# /// 13. `[writable]` quote_node_bank_ai - EntropyGroup quote vault acc
# /// 14. `[writable]` base_vault_ai - EntropyGroup base vault acc
# /// 15. `[writable]` quote_vault_ai - EntropyGroup quote vault acc
# /// 16. `[]` dex_signer_ai - dex PySerumMarket signer account
# /// 17. `[]` spl token program
#
def build_spot_settle_instructions(
    context: Context,
    wallet: Wallet,
    account: IAccount,
    market: PySerumMarket,
    group: IGroup,
    open_orders_address: PublicKey,
    base_rootbank: RootBank,
    base_nodebank: NodeBank,
    quote_rootbank: RootBank,
    quote_nodebank: NodeBank,
) -> CombinableInstructions:
    vault_signer = PublicKey.create_program_address(
        [
            bytes(market.state.public_key()),
            market.state.vault_signer_nonce().to_bytes(8, byteorder="little"),
        ],
        market.state.program_id(),
    )

    settle_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.cache),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=context.serum_program_address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=market.state.public_key()
            ),
            AccountMeta(is_signer=False, is_writable=True, pubkey=open_orders_address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.signer_key),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=market.state.base_vault()
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=market.state.quote_vault()
            ),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=base_rootbank.address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=base_nodebank.address
            ),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=quote_rootbank.address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=quote_nodebank.address
            ),
            AccountMeta(is_signer=False, is_writable=True, pubkey=base_nodebank.vault),
            AccountMeta(is_signer=False, is_writable=True, pubkey=quote_nodebank.vault),
            AccountMeta(is_signer=False, is_writable=False, pubkey=vault_signer),
            AccountMeta(is_signer=False, is_writable=False, pubkey=TOKEN_PROGRAM_ID),
        ],
        program_id=context.entropy_program_address,
        data=layouts.SETTLE_FUNDS.build({}),
    )

    return CombinableInstructions(signers=[], instructions=[settle_instruction])


def build_spot_create_openorders_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    spot_market: ISpotMarket,
    open_orders_address: PublicKey,
) -> CombinableInstructions:
    instructions: CombinableInstructions = CombinableInstructions.empty()

    # /// Accounts expected by this instruction (8):
    # ///
    # /// 0. `[]` entropy_group_ai - EntropyGroup that this entropy account is for
    # /// 1. `[writable]` entropy_account_ai - EntropyAccount
    # /// 2. `[signer]` owner_ai - EntropyAccount owner
    # /// 3. `[]` dex_prog_ai - program id of serum dex
    # /// 4. `[writable]` open_orders_ai - open orders PDA
    # /// 5. `[]` spot_market_ai - dex MarketState account
    # /// 6. `[]` signer_ai - IGroup Signer IAccount
    # /// 7. `[]` system_prog_ai - System program
    create_open_orders_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=context.serum_program_address
            ),
            AccountMeta(is_signer=False, is_writable=True, pubkey=open_orders_address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=spot_market.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.signer_key),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=SYSTEM_PROGRAM_ADDRESS
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.CREATE_SPOT_OPEN_ORDERS.build({}),
    )
    instructions += CombinableInstructions(
        signers=[], instructions=[create_open_orders_instruction]
    )
    return instructions


# # 🥭 build_spot_place_order_instructions function
#
# Creates a Entropy order-placing instruction using the Serum instruction as the inner instruction. Will create
# the necessary OpenOrders account if it doesn't already exist.
#
# /// Accounts expected by PLACE_SPOT_ORDER_2 instruction (22+openorders):
# { isSigner: false, isWritable: false, pubkey: entropyGroupPk },
# { isSigner: false, isWritable: true, pubkey: entropyAccountPk },
# { isSigner: true, isWritable: false, pubkey: ownerPk },
# { isSigner: false, isWritable: false, pubkey: entropyCachePk },
# { isSigner: false, isWritable: false, pubkey: serumDexPk },
# { isSigner: false, isWritable: true, pubkey: spotMarketPk },
# { isSigner: false, isWritable: true, pubkey: bidsPk },
# { isSigner: false, isWritable: true, pubkey: asksPk },
# { isSigner: false, isWritable: true, pubkey: requestQueuePk },
# { isSigner: false, isWritable: true, pubkey: eventQueuePk },
# { isSigner: false, isWritable: true, pubkey: spotMktBaseVaultPk },
# { isSigner: false, isWritable: true, pubkey: spotMktQuoteVaultPk },
# { isSigner: false, isWritable: false, pubkey: baseRootBankPk },
# { isSigner: false, isWritable: true, pubkey: baseNodeBankPk },
# { isSigner: false, isWritable: true, pubkey: baseVaultPk },
# { isSigner: false, isWritable: false, pubkey: quoteRootBankPk },
# { isSigner: false, isWritable: true, pubkey: quoteNodeBankPk },
# { isSigner: false, isWritable: true, pubkey: quoteVaultPk },
# { isSigner: false, isWritable: false, pubkey: TOKEN_PROGRAM_ID },
# { isSigner: false, isWritable: false, pubkey: signerPk },
# { isSigner: false, isWritable: false, pubkey: dexSignerPk },
# { isSigner: false, isWritable: false, pubkey: msrmOrSrmVaultPk },
# ...openOrders.map(({ pubkey, isWritable }) => ({
#   isSigner: false,
#   isWritable,
#   pubkey,
# })),
def build_spot_place_order_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    spot_market: ISpotMarket,
    open_orders_address: PublicKey,
    order_type: OrderType,
    side: Side,
    price: Decimal,
    quantity: Decimal,
    client_id: int,
    fee_discount_address: PublicKey,
) -> CombinableInstructions:
    instructions: CombinableInstructions = CombinableInstructions.empty()

    pyserum_market: PySerumMarket = spot_market.underlying_serum_market
    serum_order_type: pyserum.enums.OrderType = order_type.to_serum()
    serum_side: pyserum.enums.Side = side.to_serum()
    intrinsic_price = pyserum_market.state.price_number_to_lots(float(price))
    max_base_quantity = pyserum_market.state.base_size_number_to_lots(float(quantity))
    max_quote_quantity = (
        pyserum_market.state.base_size_number_to_lots(float(quantity))
        * pyserum_market.state.quote_lot_size()
        * pyserum_market.state.price_number_to_lots(float(price))
    )

    base_token_banks = [
        token_bank
        for token_bank in group.base_tokens_by_index
        if token_bank is not None
        and token_bank.token.mint == pyserum_market.state.base_mint()
    ]
    if len(base_token_banks) != 1:
        raise Exception(
            f"Could not find base token info for group {group.address} - length was {len(base_token_banks)} when it should be 1."
        )
    base_token_bank = base_token_banks[0]
    quote_token_bank = group.shared_quote

    base_root_bank: RootBank = base_token_bank.ensure_root_bank(context)
    base_node_bank: NodeBank = base_root_bank.pick_node_bank(context)
    quote_root_bank: RootBank = quote_token_bank.ensure_root_bank(context)
    quote_node_bank: NodeBank = quote_root_bank.pick_node_bank(context)

    vault_signer = PublicKey.create_program_address(
        [
            bytes(spot_market.address),
            pyserum_market.state.vault_signer_nonce().to_bytes(8, byteorder="little"),
        ],
        pyserum_market.state.program_id(),
    )

    place_spot_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=context.serum_program_address
            ),
            AccountMeta(is_signer=False, is_writable=True, pubkey=spot_market.address),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=spot_market.bids_address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=spot_market.asks_address
            ),
            AccountMeta(
                is_signer=False,
                is_writable=True,
                pubkey=pyserum_market.state.request_queue(),
            ),
            AccountMeta(
                is_signer=False,
                is_writable=True,
                pubkey=spot_market.event_queue_address,
            ),
            AccountMeta(
                is_signer=False,
                is_writable=True,
                pubkey=pyserum_market.state.base_vault(),
            ),
            AccountMeta(
                is_signer=False,
                is_writable=True,
                pubkey=pyserum_market.state.quote_vault(),
            ),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=base_root_bank.address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=base_node_bank.address
            ),
            AccountMeta(is_signer=False, is_writable=True, pubkey=base_node_bank.vault),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=quote_root_bank.address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=quote_node_bank.address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=quote_node_bank.vault
            ),
            AccountMeta(is_signer=False, is_writable=False, pubkey=TOKEN_PROGRAM_ID),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.signer_key),
            AccountMeta(is_signer=False, is_writable=False, pubkey=vault_signer),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=fee_discount_address
            ),
            *list(
                [
                    AccountMeta(
                        is_signer=False,
                        is_writable=(oo_address == open_orders_address),
                        pubkey=oo_address,
                    )
                    for oo_address in account.spot_open_orders
                ]
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.PLACE_SPOT_ORDER_2.build(
            dict(
                side=serum_side,
                limit_price=intrinsic_price,
                max_base_quantity=max_base_quantity,
                max_quote_quantity=max_quote_quantity,
                self_trade_behavior=pyserum.enums.SelfTradeBehavior.DECREMENT_TAKE,
                order_type=serum_order_type,
                client_id=client_id,
                limit=65535,
            )
        ),
    )

    return instructions + CombinableInstructions(
        signers=[], instructions=[place_spot_instruction]
    )


# # 🥭 build_spot_cancel_order_instructions function
#
# Builds the instructions necessary for cancelling a spot order.
#
def build_spot_cancel_order_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    market: PySerumMarket,
    order: Order,
    open_orders_address: PublicKey,
) -> CombinableInstructions:
    # { buy: 0, sell: 1 }
    raw_side: int = 1 if order.side == Side.SELL else 0

    # Accounts expected by this instruction:
    # { isSigner: false, isWritable: false, pubkey: entropyGroupPk },
    # { isSigner: true, isWritable: false, pubkey: ownerPk },
    # { isSigner: false, isWritable: false, pubkey: entropyAccountPk },
    # { isSigner: false, isWritable: false, pubkey: dexProgramId },
    # { isSigner: false, isWritable: true, pubkey: spotMarketPk },
    # { isSigner: false, isWritable: true, pubkey: bidsPk },
    # { isSigner: false, isWritable: true, pubkey: asksPk },
    # { isSigner: false, isWritable: true, pubkey: openOrdersPk },
    # { isSigner: false, isWritable: false, pubkey: signerKey },
    # { isSigner: false, isWritable: true, pubkey: eventQueuePk },

    instructions = [
        TransactionInstruction(
            keys=[
                AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
                AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
                AccountMeta(is_signer=False, is_writable=False, pubkey=account.address),
                AccountMeta(
                    is_signer=False,
                    is_writable=False,
                    pubkey=context.serum_program_address,
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=market.state.public_key()
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=market.state.bids()
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=market.state.asks()
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=open_orders_address
                ),
                AccountMeta(
                    is_signer=False, is_writable=False, pubkey=group.signer_key
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=market.state.event_queue()
                ),
            ],
            program_id=context.entropy_program_address,
            data=layouts.CANCEL_SPOT_ORDER.build(
                {"order_id": order.id, "side": raw_side}
            ),
        )
    ]
    return CombinableInstructions(signers=[], instructions=instructions)


# #
# # 🥭 Perp instruction builders
# #


# # 🥭 build_perp_cancel_order_instructions function
#
# Builds the instructions necessary for cancelling a perp order.
#
def build_perp_cancel_order_instructions(
    context: Context,
    wallet: Wallet,
    account: IAccount,
    perp_market_details: PerpMarketDetails,
    order: Order,
    invalid_id_ok: bool,
) -> CombinableInstructions:
    # Prefer cancelling by client ID so we don't have to keep track of the order side.
    if order.client_id is not None and order.client_id != 0:
        data: bytes = layouts.CANCEL_PERP_ORDER_BY_CLIENT_ID.build(
            {"client_order_id": order.client_id, "invalid_id_ok": invalid_id_ok}
        )
    else:
        data = layouts.CANCEL_PERP_ORDER.build(
            {"order_id": order.id, "invalid_id_ok": invalid_id_ok}
        )

    # Accounts expected by this instruction (both CANCEL_PERP_ORDER and CANCEL_PERP_ORDER_BY_CLIENT_ID are the same):
    # { isSigner: false, isWritable: false, pubkey: entropyGroupPk },
    # { isSigner: false, isWritable: true, pubkey: entropyAccountPk },
    # { isSigner: true, isWritable: false, pubkey: ownerPk },
    # { isSigner: false, isWritable: true, pubkey: perpMarketPk },
    # { isSigner: false, isWritable: true, pubkey: bidsPk },
    # { isSigner: false, isWritable: true, pubkey: asksPk },

    instructions = [
        TransactionInstruction(
            keys=[
                AccountMeta(
                    is_signer=False, is_writable=False, pubkey=account.group_address
                ),
                AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
                AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
                AccountMeta(
                    is_signer=False,
                    is_writable=True,
                    pubkey=perp_market_details.address,
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=perp_market_details.bids
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=perp_market_details.asks
                ),
            ],
            program_id=context.entropy_program_address,
            data=data,
        )
    ]
    return CombinableInstructions(signers=[], instructions=instructions)


# # 🥭 build_perp_cancel_all_orders_instructions function
#
# Builds the instructions necessary for cancelling all perp orders.
#
def build_perp_cancel_all_orders_instructions(
    context: Context,
    wallet: Wallet,
    account: IAccount,
    perp_market_details: PerpMarketDetails,
    limit: Decimal = Decimal(32),
) -> CombinableInstructions:
    # Accounts expected by this instruction (seems to be the same as CANCEL_PERP_ORDER and CANCEL_PERP_ORDER_BY_CLIENT_ID):
    # { isSigner: false, isWritable: false, pubkey: entropyGroupPk },
    # { isSigner: false, isWritable: true, pubkey: entropyAccountPk },
    # { isSigner: true, isWritable: false, pubkey: ownerPk },
    # { isSigner: false, isWritable: true, pubkey: perpMarketPk },
    # { isSigner: false, isWritable: true, pubkey: bidsPk },
    # { isSigner: false, isWritable: true, pubkey: asksPk },
    instructions = [
        TransactionInstruction(
            keys=[
                AccountMeta(
                    is_signer=False, is_writable=False, pubkey=account.group_address
                ),
                AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
                AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
                AccountMeta(
                    is_signer=False,
                    is_writable=True,
                    pubkey=perp_market_details.address,
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=perp_market_details.bids
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=perp_market_details.asks
                ),
            ],
            program_id=context.entropy_program_address,
            data=layouts.CANCEL_ALL_PERP_ORDERS.build({"limit": limit}),
        )
    ]
    return CombinableInstructions(signers=[], instructions=instructions)


# # 🥭 build_perp_cancel_all_side_instructions function
#
# Builds the instructions necessary for cancelling all perp orders on a given side.
#
def build_perp_cancel_all_side_instructions(
    context: Context,
    wallet: Wallet,
    account: IAccount,
    perp_market_details: PerpMarketDetails,
    side: Side,
    limit: Decimal = Decimal(32),
) -> CombinableInstructions:
    # { buy: 0, sell: 1 }
    raw_side: int = 1 if side == Side.SELL else 0

    # /// Cancel all perp open orders for one side of the book
    # ///
    # /// Accounts expected: 6
    # /// 0. `[]` entropy_group_ai - EntropyGroup
    # /// 1. `[writable]` entropy_account_ai - EntropyAccount
    # /// 2. `[signer]` owner_ai - Owner of Entropy Account
    # /// 3. `[writable]` perp_market_ai - PerpMarket
    # /// 4. `[writable]` bids_ai - Bids acc
    # /// 5. `[writable]` asks_ai - Asks acc
    instructions = [
        TransactionInstruction(
            keys=[
                AccountMeta(
                    is_signer=False, is_writable=False, pubkey=account.group_address
                ),
                AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
                AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
                AccountMeta(
                    is_signer=False,
                    is_writable=True,
                    pubkey=perp_market_details.address,
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=perp_market_details.bids
                ),
                AccountMeta(
                    is_signer=False, is_writable=True, pubkey=perp_market_details.asks
                ),
            ],
            program_id=context.entropy_program_address,
            data=layouts.CANCEL_PERP_ORDER_SIDE.build(
                {"side": raw_side, "limit": limit}
            ),
        )
    ]
    return CombinableInstructions(signers=[], instructions=instructions)


# /// Place an order on a perp market
# ///
# /// In case this order is matched, the corresponding order structs on both
# /// PerpAccounts (taker & maker) will be adjusted, and the position size
# /// will be adjusted w/o accounting for fees.
# /// In addition a FillEvent will be placed on the event queue.
# /// Through a subsequent invocation of ConsumeEvents the FillEvent can be
# /// executed and the perp account balances (base/quote) and fees will be
# /// paid from the quote position. Only at this point the position balance
# /// is 100% reflecting the trade.
# ///
# /// Accounts expected by this instruction (9 + `NUM_IN_MARGIN_BASKET`):
# /// 0. `[]` entropy_group_ai - EntropyGroup
# /// 1. `[writable]` entropy_account_ai - the EntropyAccount of owner
# /// 2. `[signer]` owner_ai - owner of EntropyAccount
# /// 3. `[]` entropy_cache_ai - EntropyCache for this EntropyGroup
# /// 4. `[writable]` perp_market_ai
# /// 5. `[writable]` bids_ai - bids account for this PerpMarket
# /// 6. `[writable]` asks_ai - asks account for this PerpMarket
# /// 7. `[writable]` event_queue_ai - EventQueue for this PerpMarket
# /// 8. `[writable]` referrer_entropy_account_ai - referrer's entropy account;
# ///                 pass in entropy_account_ai as duplicate if you don't have a referrer
# /// 9..9 + NUM_IN_MARGIN_BASKET `[]` open_orders_ais - pass in open orders in margin basket
def build_perp_place_order_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    perp_market_details: PerpMarketDetails,
    price: Decimal,
    quantity: Decimal,
    client_order_id: int,
    side: Side,
    order_type: OrderType,
    reduce_only: bool = False,
    expiration: datetime = Order.NoExpiration,
    match_limit: int = 20,
    max_quote_quantity: Decimal = Decimal(0),
    reflink: typing.Optional[PublicKey] = None,
) -> CombinableInstructions:
    # { buy: 0, sell: 1 }
    raw_side: int = 1 if side == Side.SELL else 0
    raw_order_type: int = order_type.to_perp()

    base_decimals = perp_market_details.base_instrument.decimals
    quote_decimals = perp_market_details.quote_token.token.decimals

    base_factor = Decimal(10) ** base_decimals
    quote_factor = Decimal(10) ** quote_decimals

    native_price = ((price * quote_factor) * perp_market_details.base_lot_size) / (
        perp_market_details.quote_lot_size * base_factor
    )
    native_quantity = (quantity * base_factor) / perp_market_details.base_lot_size
    native_max_quote_quantity = (
        (max_quote_quantity * quote_factor) / perp_market_details.quote_lot_size
    ) or I64_MAX

    # /// Accounts expected by this instruction (9 + `NUM_IN_MARGIN_BASKET`):
    # /// 0. `[]` entropy_group_ai - EntropyGroup
    # /// 1. `[writable]` entropy_account_ai - the EntropyAccount of owner
    # /// 2. `[signer]` owner_ai - owner of EntropyAccount
    # /// 3. `[]` entropy_cache_ai - EntropyCache for this EntropyGroup
    # /// 4. `[writable]` perp_market_ai
    # /// 5. `[writable]` bids_ai - bids account for this PerpMarket
    # /// 6. `[writable]` asks_ai - asks account for this PerpMarket
    # /// 7. `[writable]` event_queue_ai - EventQueue for this PerpMarket
    # /// 8. `[writable]` referrer_entropy_account_ai - referrer's entropy account;
    # ///                 pass in entropy_account_ai as duplicate if you don't have a referrer
    # /// 9..9 + NUM_IN_MARGIN_BASKET `[]` open_orders_ais - pass in open orders in margin basket
    keys = [
        AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
        AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
        AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
        AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
        AccountMeta(
            is_signer=False, is_writable=True, pubkey=perp_market_details.address
        ),
        AccountMeta(is_signer=False, is_writable=True, pubkey=perp_market_details.bids),
        AccountMeta(is_signer=False, is_writable=True, pubkey=perp_market_details.asks),
        AccountMeta(
            is_signer=False, is_writable=True, pubkey=perp_market_details.event_queue
        ),
        AccountMeta(
            is_signer=False, is_writable=True, pubkey=reflink or account.address
        ),
        *list(
            [
                AccountMeta(
                    is_signer=False,
                    is_writable=False,
                    pubkey=oo_address,
                )
                for oo_address in account.spot_open_orders
            ]
        ),
    ]

    instructions = [
        TransactionInstruction(
            keys=keys,
            program_id=context.entropy_program_address,
            data=layouts.PLACE_PERP_ORDER_2.build(
                {
                    "price": native_price,
                    "max_base_quantity": native_quantity,
                    "max_quote_quantity": native_max_quote_quantity,
                    "client_order_id": client_order_id,
                    "expiry_timestamp": expiration,
                    "side": raw_side,
                    "order_type": raw_order_type,
                    "reduce_only": reduce_only,
                    "limit": match_limit,
                }
            ),
        )
    ]

    return CombinableInstructions(signers=[], instructions=instructions)


def build_perp_consume_events_instructions(
    context: Context,
    group: IGroup,
    perp_market_details: PerpMarketDetails,
    account_addresses: typing.Sequence[PublicKey],
    limit: Decimal = Decimal(32),
) -> CombinableInstructions:
    # Accounts expected by this instruction:
    # { isSigner: false, isWritable: false, pubkey: entropyGroupPk },
    # { isSigner: false, isWritable: false, pubkey: entropyCachePk },
    # { isSigner: false, isWritable: true, pubkey: perpMarketPk },
    # { isSigner: false, isWritable: true, pubkey: eventQueuePk },
    # ...entropyAccountPks.sort().map((pubkey) => ({
    #     isSigner: false,
    #     isWritable: true,
    #     pubkey,
    # })),

    instructions = [
        TransactionInstruction(
            keys=[
                AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
                AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
                AccountMeta(
                    is_signer=False,
                    is_writable=True,
                    pubkey=perp_market_details.address,
                ),
                AccountMeta(
                    is_signer=False,
                    is_writable=True,
                    pubkey=perp_market_details.event_queue,
                ),
                *list(
                    [
                        AccountMeta(
                            is_signer=False, is_writable=True, pubkey=account_address
                        )
                        for account_address in account_addresses
                    ]
                ),
            ],
            program_id=context.entropy_program_address,
            data=layouts.CONSUME_EVENTS.build(
                {
                    "limit": limit,
                }
            ),
        )
    ]
    return CombinableInstructions(signers=[], instructions=instructions)


# #
# # 🥭 Entropy instruction builders
# #


# The old INIT_MANGO_ACCOUNT instruction is now superseded by CREATE_MANGO_ACCOUNT
def build_entropy_create_account_instructions(
    context: Context, wallet: Wallet, group: IGroup, account_num: Decimal = Decimal(1)
) -> CombinableInstructions:
    entropy_account_address_and_nonce: typing.Tuple[
        PublicKey, int
    ] = PublicKey.find_program_address(
        [
            bytes(group.address),
            bytes(wallet.address),
            int(account_num).to_bytes(8, "little"),
        ],
        context.entropy_program_address,
    )
    entropy_account_address: PublicKey = entropy_account_address_and_nonce[0]

    # /// 0. `[writable]` entropy_group_ai - EntropyGroup that this entropy account is for
    # /// 1. `[writable]` entropy_account_ai - the entropy account data
    # /// 2. `[signer]` owner_ai - Solana account of owner of the entropy account
    # /// 3. `[]` system_prog_ai - System program
    create = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.address),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=entropy_account_address
            ),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=SYSTEM_PROGRAM_ADDRESS
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.CREATE_MANGO_ACCOUNT.build({"account_num": account_num}),
    )
    return CombinableInstructions(signers=[], instructions=[create])


# /// Deposit funds into entropy account
# ///
# /// Accounts expected by this instruction (8):
# ///
# /// 0. `[]` entropy_group_ai - EntropyGroup that this entropy account is for
# /// 1. `[writable]` entropy_account_ai - the entropy account for this user
# /// 2. `[signer]` owner_ai - Solana account of owner of the entropy account
# /// 3. `[]` entropy_cache_ai - EntropyCache
# /// 4. `[]` root_bank_ai - RootBank owned by EntropyGroup
# /// 5. `[writable]` node_bank_ai - NodeBank owned by RootBank
# /// 6. `[writable]` vault_ai - TokenAccount owned by EntropyGroup
# /// 7. `[]` token_prog_ai - acc pointed to by SPL token program id
# /// 8. `[writable]` owner_token_account_ai - TokenAccount owned by user which will be sending the funds
# Deposit {
#     quantity: u64,
# },
def build_entropy_deposit_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    root_bank: RootBank,
    node_bank: NodeBank,
    token_account: TokenAccount,
) -> CombinableInstructions:
    value = token_account.value.shift_to_native().value
    deposit = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
            AccountMeta(is_signer=False, is_writable=False, pubkey=root_bank.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.vault),
            AccountMeta(is_signer=False, is_writable=False, pubkey=TOKEN_PROGRAM_ID),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=token_account.address
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.DEPOSIT.build({"quantity": value}),
    )

    return CombinableInstructions(signers=[], instructions=[deposit])


# /// Withdraw funds that were deposited earlier.
# ///
# /// Accounts expected by this instruction (10):
# ///
# /// 0. `[read]` entropy_group_ai,   -
# /// 1. `[write]` entropy_account_ai, -
# /// 2. `[read]` owner_ai,         -
# /// 3. `[read]` entropy_cache_ai,   -
# /// 4. `[read]` root_bank_ai,     -
# /// 5. `[write]` node_bank_ai,     -
# /// 6. `[write]` vault_ai,         -
# /// 7. `[write]` token_account_ai, -
# /// 8. `[read]` signer_ai,        -
# /// 9. `[read]` token_prog_ai,    -
# /// 10. `[read]` clock_ai,         -
# /// 11..+ `[]` open_orders_accs - open orders for each of the spot market
# Withdraw {
#     quantity: u64,
#     allow_borrow: bool,
# },
def build_entropy_withdraw_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    root_bank: RootBank,
    node_bank: NodeBank,
    token_account: TokenAccount,
    allow_borrow: bool,
) -> CombinableInstructions:
    value = token_account.value.shift_to_native().value
    withdraw = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
            AccountMeta(is_signer=False, is_writable=False, pubkey=root_bank.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.vault),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=token_account.address
            ),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.signer_key),
            AccountMeta(is_signer=False, is_writable=False, pubkey=TOKEN_PROGRAM_ID),
            *list(
                [
                    AccountMeta(
                        is_signer=False,
                        is_writable=False,
                        pubkey=oo_address or SYSTEM_PROGRAM_ADDRESS,
                    )
                    for oo_address in account.spot_open_orders_by_index
                ]
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.WITHDRAW.build({"quantity": value, "allow_borrow": allow_borrow}),
    )

    return CombinableInstructions(signers=[], instructions=[withdraw])


# # 🥭 build_entropy_redeem_accrued_instructions function
#
# Creates a 'RedeemMngo' instruction for Entropy accounts.
#
def build_entropy_redeem_accrued_instructions(
    context: Context,
    wallet: Wallet,
    perp_market: IPerpMarket,
    group: IGroup,
    account: IAccount,
    mngo: TokenBank,
) -> CombinableInstructions:
    node_bank: NodeBank = mngo.pick_node_bank(context)
    # /// Redeem the mngo_accrued in a PerpAccount for MNGO in EntropyAccount deposits
    # ///
    # /// Accounts expected by this instruction (11):
    # /// 0. `[]` entropy_group_ai - EntropyGroup that this entropy account is for
    # /// 1. `[]` entropy_cache_ai - EntropyCache
    # /// 2. `[writable]` entropy_account_ai - EntropyAccount
    # /// 3. `[signer]` owner_ai - EntropyAccount owner
    # /// 4. `[]` perp_market_ai - IPerpMarket
    # /// 5. `[writable]` mngo_perp_vault_ai
    # /// 6. `[]` mngo_root_bank_ai
    # /// 7. `[writable]` mngo_node_bank_ai
    # /// 8. `[writable]` mngo_bank_vault_ai
    # /// 9. `[]` signer_ai - IGroup Signer IAccount
    # /// 10. `[]` token_prog_ai - SPL Token program id
    redeem_accrued_entropy_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=perp_market.address),
            AccountMeta(
                is_signer=False,
                is_writable=True,
                pubkey=perp_market.underlying_perp_market.mngo_vault,
            ),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=mngo.root_bank_address
            ),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.vault),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.signer_key),
            AccountMeta(is_signer=False, is_writable=False, pubkey=TOKEN_PROGRAM_ID),
        ],
        program_id=context.entropy_program_address,
        data=layouts.REDEEM_MNGO.build({}),
    )
    return CombinableInstructions(
        signers=[], instructions=[redeem_accrued_entropy_instruction]
    )


# # 🥭 build_entropy_set_account_delegate_instructions function
#
# Creates an instruction to delegate account operations (except Withdraw and CloseAccount) to a
# different account.
#
# Set to SYSTEM_PROGRAM_ADDRESS to revoke delegate.
#
def build_entropy_set_account_delegate_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    delegate: PublicKey,
) -> CombinableInstructions:
    # /// https://github.com/blockworks-foundation/entropy-v3/pull/97/
    # /// Set delegate authority to entropy account which can do everything regular account can do
    # /// except Withdraw and CloseEntropyAccount. Set to Pubkey::default() to revoke delegate
    # ///
    # /// Accounts expected: 4
    # /// 0. `[]` entropy_group_ai - EntropyGroup
    # /// 1. `[writable]` entropy_account_ai - EntropyAccount
    # /// 2. `[signer]` owner_ai - Owner of Entropy IAccount
    # /// 3. `[]` delegate_ai - delegate
    set_delegate_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=delegate),
        ],
        program_id=context.entropy_program_address,
        data=layouts.SET_DELEGATE.build({}),
    )
    return CombinableInstructions(signers=[], instructions=[set_delegate_instruction])


def build_entropy_unset_account_delegate_instructions(
    context: Context, wallet: Wallet, group: IGroup, account: IAccount
) -> CombinableInstructions:
    return build_entropy_set_account_delegate_instructions(
        context, wallet, group, account, SYSTEM_PROGRAM_ADDRESS
    )


# # 🥭 build_entropy_set_referrer_memory_instructions function
#
# Creates an instruction to store the referrer's EntropyAccount pubkey on the Referrer account
# and create the Referrer account as a PDA of user's EntropyAccount if it doesn't exist
#
def build_entropy_set_referrer_memory_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    referrer_memory_address: PublicKey,
    referrer_account_address: PublicKey,
) -> CombinableInstructions:
    # /// Store the referrer's EntropyAccount pubkey on the Referrer account
    # /// It will create the Referrer account as a PDA of user's EntropyAccount if it doesn't exist
    # /// This is primarily useful for the UI; the referrer address stored here is not necessarily
    # /// who earns the ref fees.
    # ///
    # /// Accounts expected by this instruction (7):
    # ///
    # /// 0. `[]` entropy_group_ai - EntropyGroup that this entropy account is for
    # /// 1. `[]` entropy_account_ai - EntropyAccount of the referred
    # /// 2. `[signer]` owner_ai - EntropyAccount owner or delegate
    # /// 3. `[writable]` referrer_memory_ai - ReferrerMemory struct; will be initialized if required
    # /// 4. `[]` referrer_entropy_account_ai - referrer's EntropyAccount
    # /// 5. `[signer, writable]` payer_ai - payer for PDA; can be same as owner
    # /// 6. `[]` system_prog_ai - System program
    set_referrer_memory_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=account.address),
            AccountMeta(is_signer=True, is_writable=False, pubkey=wallet.address),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=referrer_memory_address
            ),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=referrer_account_address
            ),
            AccountMeta(is_signer=True, is_writable=True, pubkey=wallet.address),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=SYSTEM_PROGRAM_ADDRESS
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.SET_REFERRER_MEMORY.build({}),
    )
    return CombinableInstructions(
        signers=[], instructions=[set_referrer_memory_instruction]
    )


# # 🥭 build_entropy_register_referrer_id_instructions function
#
# Creates an instruction to register a 'referrer ID' for a Entropy IAccount
#
def build_entropy_register_referrer_id_instructions(
    context: Context,
    wallet: Wallet,
    group: IGroup,
    account: IAccount,
    referrer_record_address: PublicKey,
    referrer_id: str,
) -> CombinableInstructions:
    # /// Associate the referrer's EntropyAccount with a human readable `referrer_id` which can be used
    # /// in a ref link. This is primarily useful for the UI.
    # /// Create the `ReferrerIdRecord` PDA; if it already exists throw error
    # ///
    # /// Accounts expected by this instruction (5):
    # /// 0. `[]` entropy_group_ai - EntropyGroup
    # /// 1. `[]` referrer_entropy_account_ai - EntropyAccount
    # /// 2. `[writable]` referrer_id_record_ai - The PDA to store the record on
    # /// 3. `[signer, writable]` payer_ai - payer for PDA; can be same as owner
    # /// 4. `[]` system_prog_ai - System program
    register_referrer_id_instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=account.address),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=referrer_record_address
            ),
            AccountMeta(is_signer=True, is_writable=True, pubkey=wallet.address),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=SYSTEM_PROGRAM_ADDRESS
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.REGISTER_REFERRER_ID.build({"info": referrer_id}),
    )
    return CombinableInstructions(
        signers=[], instructions=[register_referrer_id_instruction]
    )


def build_entropy_cache_root_banks_instructions(
    context: Context,
    group: IGroup,
    root_banks: typing.Sequence[PublicKey],
) -> CombinableInstructions:
    # /// DEPRECATED - caching of root banks now happens in update index
    # /// Cache root banks
    # ///
    # /// Accounts expected: 2 + Root Banks
    # /// 0. `[]` entropy_group_ai
    # /// 1. `[writable]` entropy_cache_ai
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.cache),
            *list(
                [
                    AccountMeta(
                        is_signer=False,
                        is_writable=False,
                        pubkey=root_bank,
                    )
                    for root_bank in root_banks
                ]
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.CACHE_ROOT_BANKS.build({}),
    )
    return CombinableInstructions(signers=[], instructions=[instruction])


def build_entropy_cache_prices_instructions(
    context: Context,
    group: IGroup,
    oracle_addresses: typing.Sequence[PublicKey],
) -> CombinableInstructions:
    # /// Cache prices
    # ///
    # /// Accounts expected: 3 + Oracles
    # /// 0. `[]` entropy_group_ai -
    # /// 1. `[writable]` entropy_cache_ai -
    # /// 2+... `[]` oracle_ais - flux aggregator feed accounts
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.cache),
            *list(
                [
                    AccountMeta(
                        is_signer=False,
                        is_writable=False,
                        pubkey=oracle,
                    )
                    for oracle in oracle_addresses
                ]
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.CACHE_PRICES.build({}),
    )
    return CombinableInstructions(signers=[], instructions=[instruction])


def build_entropy_cache_perp_markets_instructions(
    context: Context,
    group: IGroup,
    perp_market_addresses: typing.Sequence[PublicKey],
) -> CombinableInstructions:
    # /// Cache perp markets
    # ///
    # /// Accounts expected: 2 + Perp Markets
    # /// 0. `[]` entropy_group_ai
    # /// 1. `[writable]` entropy_cache_ai
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.cache),
            *map(
                lambda address: AccountMeta(
                    is_signer=False,
                    is_writable=False,
                    pubkey=address,
                ),
                perp_market_addresses,
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.CACHE_PERP_MARKETS.build({}),
    )
    return CombinableInstructions(signers=[], instructions=[instruction])


def build_entropy_update_root_bank_instructions(
    context: Context,
    group: IGroup,
    root_bank: PublicKey,
    node_banks: typing.Sequence[PublicKey],
) -> CombinableInstructions:
    # /// Update a root bank's indexes by providing all it's node banks
    # ///
    # /// Accounts expected: 2 + Node Banks
    # /// 0. `[]` entropy_group_ai - EntropyGroup
    # /// 1. `[]` root_bank_ai - RootBank
    # /// 2+... `[]` node_bank_ais - NodeBanks
    #
    # Note: The above doesn't seem quite right. It seems to need the Cache as [1], and
    # cache+banks all seem to need to be writable.
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.cache),
            AccountMeta(is_signer=False, is_writable=True, pubkey=root_bank),
            *list(
                [
                    AccountMeta(
                        is_signer=False,
                        is_writable=True,
                        pubkey=node_bank,
                    )
                    for node_bank in node_banks
                ]
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.UPDATE_ROOT_BANK.build({}),
    )
    return CombinableInstructions(signers=[], instructions=[instruction])


def build_entropy_update_funding_instructions(
    context: Context,
    group: IGroup,
    perp_market_details: PerpMarketDetails,
) -> CombinableInstructions:
    # /// Update funding related variables
    #
    # Seems to take:
    #   const keys = [
    #     { isSigner: false, isWritable: false, pubkey: entropyGroupPk },
    #     { isSigner: false, isWritable: true, pubkey: entropyCachePk },
    #     { isSigner: false, isWritable: true, pubkey: perpMarketPk },
    #     { isSigner: false, isWritable: false, pubkey: bidsPk },
    #     { isSigner: false, isWritable: false, pubkey: asksPk },
    #   ];
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.cache),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=perp_market_details.address
            ),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=perp_market_details.bids
            ),
            AccountMeta(
                is_signer=False, is_writable=False, pubkey=perp_market_details.asks
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.UPDATE_FUNDING.build({}),
    )
    return CombinableInstructions(signers=[], instructions=[instruction])


def build_entropy_settle_fees_instructions(
    context: Context,
    group: IGroup,
    perp_market_details: PerpMarketDetails,
    account: IAccount,
    root_bank: RootBank,
    node_bank: NodeBank,
) -> CombinableInstructions:
    # /// Take an account that has losses in the selected perp market to account for fees_accrued
    # ///
    # /// Accounts expected: 10
    # /// 0. `[]` entropy_group_ai - EntropyGroup
    # /// 1. `[]` entropy_cache_ai - EntropyCache
    # /// 2. `[writable]` perp_market_ai - PerpMarket
    # /// 3. `[writable]` entropy_account_ai - EntropyAccount
    # /// 4. `[]` root_bank_ai - RootBank
    # /// 5. `[writable]` node_bank_ai - NodeBank
    # /// 6. `[writable]` bank_vault_ai - ?
    # /// 7. `[writable]` fees_vault_ai - fee vault owned by entropy DAO token governance
    # /// 8. `[]` signer_ai - Group Signer Account
    # /// 9. `[]` token_prog_ai - Token Program Account
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=perp_market_details.address
            ),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=root_bank.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=node_bank.vault),
            AccountMeta(is_signer=False, is_writable=True, pubkey=group.fees_vault),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.signer_key),
            AccountMeta(is_signer=False, is_writable=False, pubkey=TOKEN_PROGRAM_ID),
        ],
        program_id=context.entropy_program_address,
        data=layouts.SETTLE_FEES.build({}),
    )
    return CombinableInstructions(signers=[], instructions=[instruction])


def build_entropy_settle_pnl_instructions(
    context: Context,
    group: IGroup,
    group_slot: IGroupSlot,
    account_a: IAccount,
    account_b: IAccount,
    root_bank: RootBank,
) -> CombinableInstructions:
    # /// Take two EntropyAccounts and settle profits and losses between them for a perp market
    # ///
    # /// Accounts expected (6):
    #   const keys = [
    #     { isSigner: false, isWritable: false, pubkey: entropyGroupPk },
    #     { isSigner: false, isWritable: true, pubkey: entropyAccountAPk },
    #     { isSigner: false, isWritable: true, pubkey: entropyAccountBPk },
    #     { isSigner: false, isWritable: false, pubkey: entropyCachePk },
    #     { isSigner: false, isWritable: false, pubkey: rootBankPk },
    #     { isSigner: false, isWritable: true, pubkey: nodeBankPk },
    #   ];
    instruction = TransactionInstruction(
        keys=[
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account_a.address),
            AccountMeta(is_signer=False, is_writable=True, pubkey=account_b.address),
            AccountMeta(is_signer=False, is_writable=False, pubkey=group.cache),
            AccountMeta(is_signer=False, is_writable=False, pubkey=root_bank.address),
            AccountMeta(
                is_signer=False, is_writable=True, pubkey=root_bank.node_banks[0]
            ),
        ],
        program_id=context.entropy_program_address,
        data=layouts.SETTLE_PNL.build({"market_index": group_slot.index}),
    )
    return CombinableInstructions(signers=[], instructions=[instruction])



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import typing


def indent_collection_as_str(
    collection: typing.Sequence[typing.Any], levels: int = 0
) -> str:
    spaces: int = levels * 4
    spacing: str = " " * spaces
    if len(collection) == 0:
        return f"{spacing}None"
    return f"{spacing}" + f"\n{spacing}".join(f"{item}" for item in collection)


def indent_item_by(item: typing.Any, levels: int = 1) -> str:
    spaces: int = levels * 4
    spacing: str = " " * spaces
    return f"{item}".replace("\n", f"\n{spacing}")



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://org-docs.friktion.fi/entropy/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/Friktion-Labs)
#   [Email](mailto:hello@blockworks.foundation)

import rx.operators
import typing

from datetime import datetime
from decimal import Decimal
from pyserum.market.market import Market as PySerumMarket
from pyserum.market.orderbook import OrderBook as PySerumOrderBook
from solana.publickey import PublicKey

from entropy.datetimes import utc_now

from .accountinfo import AccountInfo
from .combinableinstructions import CombinableInstructions
from .constants import SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .instructions import (
    build_serum_consume_events_instructions,
    build_serum_create_openorders_instructions,
    build_serum_settle_instructions,
    build_serum_place_order_instructions,
)
from .loadedmarket import LoadedMarket
from .lotsizeconverter import LotSizeConverter, RaisingLotSizeConverter
from .markets import InventorySource, Market, MarketType
from .marketoperations import MarketInstructionBuilder, MarketOperations
from .openorders import OpenOrders
from .observables import Disposable
from .orders import Order, OrderBook, Side
from .publickey import encode_public_key_for_sorting
from .serumeventqueue import SerumEvent, SerumEventQueue, UnseenSerumEventChangesTracker
from .tokens import Instrument, Token
from .tokenaccount import TokenAccount
from .wallet import Wallet
from .websocketsubscription import (
    IndividualWebSocketSubscriptionManager,
    WebSocketAccountSubscription,
)


# # 🥭 SerumMarket class
#
# This class encapsulates our knowledge of a Serum spot market.
#
class SerumMarket(LoadedMarket):
    def __init__(
        self,
        serum_program_address: PublicKey,
        address: PublicKey,
        base: Token,
        quote: Token,
        underlying_serum_market: PySerumMarket,
    ) -> None:
        super().__init__(
            MarketType.SERUM,
            serum_program_address,
            address,
            InventorySource.SPL_TOKENS,
            base,
            quote,
            RaisingLotSizeConverter(),
        )
        self.base: Token = base
        self.quote: Token = quote
        self.underlying_serum_market: PySerumMarket = underlying_serum_market
        base_lot_size: Decimal = Decimal(underlying_serum_market.state.base_lot_size())
        quote_lot_size: Decimal = Decimal(
            underlying_serum_market.state.quote_lot_size()
        )
        self.lot_size_converter: LotSizeConverter = LotSizeConverter(
            base, base_lot_size, quote, quote_lot_size
        )

    @staticmethod
    def isa(market: Market) -> bool:
        return market.type == MarketType.SERUM

    @staticmethod
    def ensure(market: Market) -> "SerumMarket":
        if not SerumMarket.isa(market):
            raise Exception(
                f"Market for {market.fully_qualified_symbol} is not a Serum market"
            )
        return typing.cast(SerumMarket, market)

    @property
    def fully_qualified_symbol(self) -> str:
        return f"serum:{self.symbol}"

    @property
    def bids_address(self) -> PublicKey:
        return self.underlying_serum_market.state.bids()

    @property
    def asks_address(self) -> PublicKey:
        return self.underlying_serum_market.state.asks()

    @property
    def event_queue_address(self) -> PublicKey:
        return self.underlying_serum_market.state.event_queue()

    def parse_account_info_to_orders(
        self, account_info: AccountInfo
    ) -> typing.Sequence[Order]:
        orderbook: PySerumOrderBook = PySerumOrderBook.from_bytes(
            self.underlying_serum_market.state, account_info.data
        )
        return list(map(Order.from_serum_order, orderbook.orders()))

    def unprocessed_events(self, context: Context) -> typing.Sequence[SerumEvent]:
        event_queue: SerumEventQueue = SerumEventQueue.load(
            context, self.event_queue_address, self.base, self.quote
        )
        return event_queue.unprocessed_events

    def find_openorders_address_for_owner(
        self, context: Context, owner: PublicKey
    ) -> typing.Optional[PublicKey]:
        all_open_orders = OpenOrders.load_for_market_and_owner(
            context,
            self.address,
            owner,
            context.serum_program_address,
            self.base,
            self.quote,
        )
        if len(all_open_orders) == 0:
            return None
        return all_open_orders[0].address

    def on_fill(
        self, context: Context, handler: typing.Callable[[SerumEvent], None]
    ) -> Disposable:
        def _fill_filter(item: SerumEvent) -> None:
            if item.event_flags.fill:
                handler(item)

        return self.on_event(context, _fill_filter)

    def on_event(
        self, context: Context, handler: typing.Callable[[SerumEvent], None]
    ) -> Disposable:
        disposer = Disposable()
        event_queue_address = self.event_queue_address
        initial: SerumEventQueue = SerumEventQueue.load(
            context, self.event_queue_address, self.base, self.quote
        )

        splitter: UnseenSerumEventChangesTracker = UnseenSerumEventChangesTracker(
            initial
        )
        event_queue_subscription = WebSocketAccountSubscription(
            context,
            event_queue_address,
            lambda account_info: SerumEventQueue.parse(
                account_info, self.base, self.quote
            ),
        )
        disposer.add_disposable(event_queue_subscription)

        manager = IndividualWebSocketSubscriptionManager(context)
        disposer.add_disposable(manager)
        manager.add(event_queue_subscription)

        publisher = event_queue_subscription.publisher.pipe(
            rx.operators.flat_map(splitter.unseen)
        )

        individual_event_subscription = publisher.subscribe(on_next=handler)
        disposer.add_disposable(individual_event_subscription)

        manager.open()

        return disposer

    def __str__(self) -> str:
        return f"""« SerumMarket {self.symbol} {self.address} [{self.program_address}]
    Event Queue: {self.underlying_serum_market.state.event_queue()}
    Request Queue: {self.underlying_serum_market.state.request_queue()}
    Bids: {self.underlying_serum_market.state.bids()}
    Asks: {self.underlying_serum_market.state.asks()}
    Base: [lot size: {self.underlying_serum_market.state.base_lot_size()}] {self.underlying_serum_market.state.base_mint()}
    Quote: [lot size: {self.underlying_serum_market.state.quote_lot_size()}] {self.underlying_serum_market.state.quote_mint()}
»"""


# # 🥭 SerumMarketInstructionBuilder
#
# This file deals with building instructions for Serum markets.
#
# As a matter of policy for all InstructionBuidlers, construction and build_* methods should all work with
# existing data, requiring no fetches from Solana or other sources. All necessary data should all be loaded
# on initial setup in the `load()` method.
#
class SerumMarketInstructionBuilder(MarketInstructionBuilder):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        serum_market: SerumMarket,
        raw_market: PySerumMarket,
        base_token_account: TokenAccount,
        quote_token_account: TokenAccount,
        open_orders_address: typing.Optional[PublicKey],
        fee_discount_token_address: PublicKey,
    ) -> None:
        super().__init__()
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.serum_market: SerumMarket = serum_market
        self.raw_market: PySerumMarket = raw_market
        self.base_token_account: TokenAccount = base_token_account
        self.quote_token_account: TokenAccount = quote_token_account
        self.open_orders_address: typing.Optional[PublicKey] = open_orders_address
        self.fee_discount_token_address: PublicKey = fee_discount_token_address

    @staticmethod
    def load(
        context: Context, wallet: Wallet, serum_market: SerumMarket
    ) -> "SerumMarketInstructionBuilder":
        raw_market: PySerumMarket = PySerumMarket.load(
            context.client.compatible_client,
            serum_market.address,
            context.serum_program_address,
        )

        fee_discount_token_address: PublicKey = SYSTEM_PROGRAM_ADDRESS
        srm_instrument: typing.Optional[
            Instrument
        ] = context.instrument_lookup.find_by_symbol("SRM")
        if srm_instrument is not None:
            srm_token: Token = Token.ensure(srm_instrument)
            fee_discount_token_account = TokenAccount.fetch_largest_for_owner_and_token(
                context, wallet.address, srm_token
            )
            if fee_discount_token_account is not None:
                fee_discount_token_address = fee_discount_token_account.address

        open_orders_address: typing.Optional[PublicKey] = None
        all_open_orders = OpenOrders.load_for_market_and_owner(
            context,
            serum_market.address,
            wallet.address,
            context.serum_program_address,
            serum_market.base,
            serum_market.quote,
        )
        if len(all_open_orders) > 0:
            open_orders_address = all_open_orders[0].address

        base_token_account = TokenAccount.fetch_largest_for_owner_and_token(
            context, wallet.address, serum_market.base
        )
        if base_token_account is None:
            raise Exception(
                f"Could not find source token account for base token {serum_market.base.symbol}."
            )

        quote_token_account = TokenAccount.fetch_largest_for_owner_and_token(
            context, wallet.address, serum_market.quote
        )
        if quote_token_account is None:
            raise Exception(
                f"Could not find source token account for quote token {serum_market.quote.symbol}."
            )

        return SerumMarketInstructionBuilder(
            context,
            wallet,
            serum_market,
            raw_market,
            base_token_account,
            quote_token_account,
            open_orders_address,
            fee_discount_token_address,
        )

    def build_cancel_order_instructions(
        self, order: Order, ok_if_missing: bool = False
    ) -> CombinableInstructions:
        # For us to cancel an order, an open_orders account must already exist (or have existed).
        if self.open_orders_address is None:
            raise Exception(
                f"Cannot cancel order with client ID {order.client_id} - no OpenOrders account."
            )

        raw_instruction = self.raw_market.make_cancel_order_by_client_id_instruction(
            self.wallet.keypair, self.open_orders_address, order.client_id
        )
        return CombinableInstructions.from_instruction(raw_instruction)

    def build_place_order_instructions(self, order: Order) -> CombinableInstructions:
        if order.reduce_only:
            self._logger.warning(
                "Ignoring reduce_only - not supported on Serum markets"
            )

        if order.expiration != Order.NoExpiration:
            self._logger.warning("Ignoring expiration - not supported on Serum markets")

        if order.match_limit != Order.DefaultMatchLimit:
            self._logger.warning(
                "Ignoring match_limit - not supported on Serum markets"
            )

        ensure_open_orders = CombinableInstructions.empty()
        if self.open_orders_address is None:
            ensure_open_orders = self.build_create_openorders_instructions()

        if self.open_orders_address is None:
            raise Exception("Failed to find or create OpenOrders address")

        payer_token_account = (
            self.quote_token_account
            if order.side == Side.BUY
            else self.base_token_account
        )
        place = build_serum_place_order_instructions(
            self.context,
            self.wallet,
            self.raw_market,
            payer_token_account.address,
            self.open_orders_address,
            order.order_type,
            order.side,
            order.price,
            order.quantity,
            order.client_id,
            self.fee_discount_token_address,
        )

        return ensure_open_orders + place

    def build_settle_instructions(self) -> CombinableInstructions:
        if self.open_orders_address is None:
            return CombinableInstructions.empty()
        return build_serum_settle_instructions(
            self.context,
            self.wallet,
            self.raw_market,
            self.open_orders_address,
            self.base_token_account.address,
            self.quote_token_account.address,
        )

    def build_crank_instructions(
        self, addresses: typing.Sequence[PublicKey], limit: Decimal = Decimal(32)
    ) -> CombinableInstructions:
        if self.open_orders_address is None:
            self._logger.debug(
                "Returning empty crank instructions - no serum OpenOrders address provided."
            )
            return CombinableInstructions.empty()

        distinct_addresses: typing.List[PublicKey] = []
        for oo in addresses:
            if oo not in distinct_addresses:
                distinct_addresses += [oo]

        if len(distinct_addresses) > limit:
            self._logger.warn(
                f"Cranking limited to {limit} of {len(distinct_addresses)} addresses waiting to be cranked."
            )

        limited_addresses = distinct_addresses[
            0 : min(int(limit), len(distinct_addresses))
        ]
        limited_addresses.sort(key=encode_public_key_for_sorting)

        self._logger.debug(
            f"About to crank {len(limited_addresses)} addresses: {limited_addresses}"
        )
        return build_serum_consume_events_instructions(
            self.context,
            self.serum_market.address,
            self.raw_market.state.event_queue(),
            limited_addresses,
            int(limit),
        )

    def build_create_openorders_instructions(self) -> CombinableInstructions:
        create_open_orders = build_serum_create_openorders_instructions(
            self.context, self.wallet, self.raw_market
        )
        self.open_orders_address = create_open_orders.signers[0].public_key
        return create_open_orders

    def build_redeem_instructions(self) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def __str__(self) -> str:
        return """« SerumMarketInstructionBuilder »"""


# # 🥭 SerumMarketOperations class
#
# This class performs standard operations on the Serum orderbook.
#
class SerumMarketOperations(MarketOperations):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        market_instruction_builder: SerumMarketInstructionBuilder,
    ) -> None:
        super().__init__(market_instruction_builder.serum_market)
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.market_instruction_builder: SerumMarketInstructionBuilder = (
            market_instruction_builder
        )

    @staticmethod
    def ensure(market_ops: MarketOperations) -> "SerumMarketOperations":
        if not isinstance(market_ops, SerumMarketOperations):
            raise Exception(
                f"MarketOperations for {market_ops.symbol} is not a SerumMarketOperations"
            )
        return market_ops

    @property
    def serum_market(self) -> SerumMarket:
        return self.market_instruction_builder.serum_market

    def cancel_order(
        self, order: Order, ok_if_missing: bool = False
    ) -> typing.Sequence[str]:
        self._logger.info(
            f"Cancelling {self.serum_market.fully_qualified_symbol} order {order}."
        )
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        cancel: CombinableInstructions = (
            self.market_instruction_builder.build_cancel_order_instructions(
                order, ok_if_missing=ok_if_missing
            )
        )
        crank: CombinableInstructions = self._build_crank()
        settle: CombinableInstructions = (
            self.market_instruction_builder.build_settle_instructions()
        )
        return (signers + cancel + crank + settle).execute(self.context)

    def place_order(
        self, order: Order, crank_limit: Decimal = Decimal(5)
    ) -> typing.Sequence[str]:
        client_id: int = order.client_id or self.context.generate_client_id()
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )

        open_orders_address = (
            self.market_instruction_builder.open_orders_address
            or SYSTEM_PROGRAM_ADDRESS
        )
        order_with_client_id: Order = Order(
            id=0,
            client_id=client_id,
            side=order.side,
            price=order.price,
            quantity=order.quantity,
            owner=open_orders_address,
            order_type=order.order_type,
        )
        self._logger.info(
            f"Placing {self.serum_market.fully_qualified_symbol} order {order_with_client_id}."
        )
        place: CombinableInstructions = (
            self.market_instruction_builder.build_place_order_instructions(
                order_with_client_id
            )
        )

        crank: CombinableInstructions = self._build_crank(crank_limit)
        settle: CombinableInstructions = (
            self.market_instruction_builder.build_settle_instructions()
        )

        return (signers + place + crank + settle).execute(self.context)

    def settle(self) -> typing.Sequence[str]:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        settle = self.market_instruction_builder.build_settle_instructions()
        return (signers + settle).execute(self.context)

    def crank(self, limit: Decimal = Decimal(32)) -> typing.Sequence[str]:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        crank = self._build_crank(limit)
        return (signers + crank).execute(self.context)

    def create_openorders(self) -> PublicKey:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        create_open_orders = (
            self.market_instruction_builder.build_create_openorders_instructions()
        )
        open_orders_address = create_open_orders.signers[0].public_key
        (signers + create_open_orders).execute(self.context)

        return open_orders_address

    def ensure_openorders(self) -> PublicKey:
        if self.market_instruction_builder.open_orders_address is not None:
            return self.market_instruction_builder.open_orders_address
        return self.create_openorders()

    def load_orderbook(self) -> OrderBook:
        return self.serum_market.fetch_orderbook(self.context)

    def load_my_orders(
        self, cutoff: typing.Optional[datetime] = utc_now()
    ) -> typing.Sequence[Order]:
        open_orders_address = self.market_instruction_builder.open_orders_address
        if not open_orders_address:
            return []

        orderbook: OrderBook = self.load_orderbook()
        return orderbook.all_orders_for_owner(open_orders_address, cutoff=cutoff)

    def _build_crank(
        self, limit: Decimal = Decimal(32), add_self: bool = False
    ) -> CombinableInstructions:
        open_orders_to_crank: typing.List[PublicKey] = []
        for event in self.serum_market.unprocessed_events(self.context):
            open_orders_to_crank += [event.public_key]
        self._logger.debug(f"open_orders_to_crank: {len(open_orders_to_crank)}")

        if add_self and self.market_instruction_builder.open_orders_address is not None:
            open_orders_to_crank += [
                self.market_instruction_builder.open_orders_address
            ]

        if len(open_orders_to_crank) == 0:
            return CombinableInstructions.empty()

        self._logger.debug(
            f"Building crank instruction with {len(open_orders_to_crank)} public keys, throttled to {limit}"
        )
        return self.market_instruction_builder.build_crank_instructions(
            open_orders_to_crank, limit
        )

    def __str__(self) -> str:
        return f"""« SerumMarketOperations [{self.serum_market.fully_qualified_symbol}] »"""


# # 🥭 SerumMarketStub class
#
# This class holds information to load a `SerumMarket` object but doesn't automatically load it.
#
class SerumMarketStub(Market):
    def __init__(
        self,
        serum_program_address: PublicKey,
        address: PublicKey,
        base: Token,
        quote: Token,
    ) -> None:
        super().__init__(
            MarketType.STUB,
            serum_program_address,
            address,
            InventorySource.SPL_TOKENS,
            base,
            quote,
            RaisingLotSizeConverter(),
        )
        self.base: Token = base
        self.quote: Token = quote

    @property
    def fully_qualified_symbol(self) -> str:
        return f"serum:{self.symbol}"

    def load(self, context: Context) -> SerumMarket:
        underlying_serum_market: PySerumMarket = PySerumMarket.load(
            context.client.compatible_client,
            self.address,
            context.serum_program_address,
        )
        return SerumMarket(
            self.program_address,
            self.address,
            self.base,
            self.quote,
            underlying_serum_market,
        )

    def __str__(self) -> str:
        return (
            f"« SerumMarketStub {self.symbol} {self.address} [{self.program_address}] »"
        )



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import typing

from decimal import Decimal

from .cache import PerpMarketCache
from .instrumentvalue import InstrumentValue
from .lotsizeconverter import LotSizeConverter
from .perpopenorders import PerpOpenOrders
from .tokens import Instrument, Token


# # 🥭 PerpAccount class
#
# Perp accounts aren't directly addressable. They exist as a sub-object of a full Entropy `Account` object.
#
class PerpAccount:
    def __init__(
        self,
        base_position: Decimal,
        quote_position: Decimal,
        long_settled_funding: Decimal,
        short_settled_funding: Decimal,
        bids_quantity: Decimal,
        asks_quantity: Decimal,
        taker_base: Decimal,
        taker_quote: Decimal,
        mngo_accrued: InstrumentValue,
        open_orders: PerpOpenOrders,
        lot_size_converter: LotSizeConverter,
        base_token_value: InstrumentValue,
        base_position_raw: Decimal,
        quote_position_raw: Decimal,
    ) -> None:
        self.base_position: Decimal = base_position
        self.quote_position: Decimal = quote_position
        self.long_settled_funding: Decimal = long_settled_funding
        self.short_settled_funding: Decimal = short_settled_funding
        self.bids_quantity: Decimal = bids_quantity
        self.asks_quantity: Decimal = asks_quantity
        self.taker_base: Decimal = taker_base
        self.taker_quote: Decimal = taker_quote
        self.mngo_accrued: InstrumentValue = mngo_accrued
        self.open_orders: PerpOpenOrders = open_orders
        self.lot_size_converter: LotSizeConverter = lot_size_converter
        self.base_token_value: InstrumentValue = base_token_value
        self.base_position_raw: Decimal = base_position_raw
        self.quote_position_raw: Decimal = quote_position_raw

    @staticmethod
    def from_layout(
        layout: typing.Any,
        base_token: Instrument,
        quote_token: Token,
        open_orders: PerpOpenOrders,
        lot_size_converter: LotSizeConverter,
        mngo_token: Token,
    ) -> "PerpAccount":
        base_position: Decimal = layout.base_position
        quote_position: Decimal = layout.quote_position
        long_settled_funding: Decimal = layout.long_settled_funding
        short_settled_funding: Decimal = layout.short_settled_funding
        bids_quantity: Decimal = layout.bids_quantity
        asks_quantity: Decimal = layout.asks_quantity
        taker_base: Decimal = layout.taker_base
        taker_quote: Decimal = layout.taker_quote
        mngo_accrued_raw: Decimal = layout.mngo_accrued
        mngo_accrued: InstrumentValue = InstrumentValue(
            mngo_token, mngo_token.shift_to_decimals(mngo_accrued_raw)
        )

        base_position_raw = (
            base_position + taker_base
        ) * lot_size_converter.base_lot_size
        base_token_value: InstrumentValue = InstrumentValue(
            base_token, base_position_raw
        )

        # quote_position_raw: Decimal = quote_token.shift_to_decimals(quote_position)
        quote_position_raw: Decimal = lot_size_converter.adjust_to_quote_decimals(
            quote_position
        )

        return PerpAccount(
            base_position,
            quote_position,
            long_settled_funding,
            short_settled_funding,
            bids_quantity,
            asks_quantity,
            taker_base,
            taker_quote,
            mngo_accrued,
            open_orders,
            lot_size_converter,
            base_token_value,
            base_position_raw,
            quote_position_raw,
        )

    @property
    def empty(self) -> bool:
        if (
            self.base_position == Decimal(0)
            and self.quote_position == Decimal(0)
            and self.long_settled_funding == Decimal(0)
            and self.short_settled_funding == Decimal(0)
            and self.mngo_accrued.value == Decimal(0)
            and self.open_orders.empty
        ):
            return True
        return False

    def unsettled_funding(self, perp_market_cache: PerpMarketCache) -> Decimal:
        base_position: Decimal = self.base_position
        unsettled: Decimal
        if base_position < 0:
            unsettled = base_position * (
                perp_market_cache.short_funding - self.short_settled_funding
            )
        else:
            unsettled = base_position * (
                perp_market_cache.long_funding - self.long_settled_funding
            )
        return -self.lot_size_converter.quote.shift_to_decimals(unsettled)

    def asset_value(
        self, perp_market_cache: PerpMarketCache, price: Decimal
    ) -> Decimal:
        base_position: Decimal = self.lot_size_converter.adjust_to_quote_decimals(
            self.base_position
        )
        value: Decimal = Decimal(0)
        if base_position > 0:
            value = base_position * self.lot_size_converter.base_lot_size * price
            value = self.lot_size_converter.quote.shift_to_decimals(value)

        quote_position: Decimal = self.lot_size_converter.quote.shift_to_decimals(
            self.quote_position
        )
        quote_position += self.unsettled_funding(perp_market_cache)
        if quote_position > 0:
            value += quote_position

        return value

    def liability_value(
        self, perp_market_cache: PerpMarketCache, price: Decimal
    ) -> Decimal:
        base_position: Decimal = self.lot_size_converter.adjust_to_quote_decimals(
            self.base_position
        )
        value: Decimal = Decimal(0)
        if base_position < 0:
            value = base_position * self.lot_size_converter.base_lot_size * price
            value = self.lot_size_converter.quote.shift_to_decimals(value)

        quote_position: Decimal = self.lot_size_converter.quote.shift_to_decimals(
            self.quote_position
        )
        quote_position += self.unsettled_funding(perp_market_cache)
        if quote_position < 0:
            value += quote_position

        return value

    def pnl(
        self, perp_market_cache: PerpMarketCache, price: InstrumentValue
    ) -> Decimal:
        base_position: Decimal = self.base_position_raw
        price_value: Decimal = self.lot_size_converter.adjust_to_quote_decimals(
            price.value
        )
        value: Decimal = base_position * price_value

        quote_position: Decimal = self.quote_position

        result = value + quote_position

        return self.lot_size_converter.quote.shift_to_decimals(
            result
        ) + self.unsettled_funding(perp_market_cache)

    def __str__(self) -> str:
        if self.empty:
            return "« PerpAccount (empty) »"
        open_orders = f"{self.open_orders}".replace("\n", "\n        ")
        return f"""« PerpAccount
    Base Position: {self.base_token_value}
    Quote Position: {self.quote_position}
    Long Settled Funding: {self.long_settled_funding}
    Short Settled Funding: {self.short_settled_funding}
    Bids Quantity: {self.bids_quantity}
    Asks Quantity: {self.asks_quantity}
    Taker Base: {self.taker_base}
    Taker Quote: {self.taker_quote}
    MNGO Accrued: {self.mngo_accrued}
    OpenOrders:
        {open_orders}
»"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

from datetime import datetime, timezone

# # 🥭 Datetimes
#
# Comparing datetime values in Python can lead to the following error:
#   TypeError: can't compare offset-naive and offset-aware datetimes
#
# To avoid that, we use these functions to consistently create the datetime in a way
# that has the correct value and that can be compared with other datetimes generated
# by these methods.
#
# However, getting a comparable UTC time using this mechanism:
#   return datetime.utcnow().astimezone(timezone.utc)
#
# works fine on Linux and Mac but fails on Windows. So we need to avoid that method and
# be careful of different platform issues.
#
# If you use these methods to generate the datetime values, you _should_ be safe.
#


def local_now() -> datetime:
    return datetime.now().astimezone()


# Getting a comparable UTC time using this line:
#   return datetime.utcnow().astimezone(timezone.utc)
# seems to cause `OSError: [Errno 22] Invalid argument` on Windows (but is fine on Mac and
# Linux). Best to avoid it and use the following instead, which seems to work consistently
# on Windows as well as Mac and Linux.
def utc_now() -> datetime:
    return datetime.utcnow().replace(tzinfo=timezone.utc)


# All timestamp values _should be_ in UTC. All we need to do is load the value as a UTC
# datetime that's comparable on all platforms, so we do this the same way as `utc_now()`.
def datetime_from_timestamp(timestamp: float) -> datetime:
    return datetime.utcfromtimestamp(timestamp).replace(tzinfo=timezone.utc)


# All on-chain datetime values are in UTC. All we need to do is load the value as a UTC
# datetime that's comparable on all platforms, so we do this the same way as `utc_now()`.
def datetime_from_chain(onchain_value: float) -> datetime:
    return datetime.utcfromtimestamp(onchain_value).replace(tzinfo=timezone.utc)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import entropy
import typing

from datetime import timedelta
from decimal import Decimal

from ..modelstate import ModelState
from .orderreconciler import OrderReconciler
from .reconciledorders import ReconciledOrders


# # 🥭 ToleranceOrderReconciler class
#
# Has a level of 'tolerance' around whether a desired order matches an existing order.
#
# There are two tolerance levels:
# * A tolerance for price matching
# * A tolderance for quantity matching
#
# Tolerances are expressed as a ratio. To match the existing value must be within +/- the tolderance
# of the desired value.
#
# Note:
# * A BUY only matches with a BUY, a SELL only matches with a SELL.
# * ID and Client ID are ignored when matching.
# * ModelState is ignored when matching.
#
class ToleranceOrderReconciler(OrderReconciler):
    def __init__(
        self,
        price_tolerance: Decimal,
        quantity_tolerance: Decimal,
        time_in_force_tolerance: timedelta,
    ) -> None:
        super().__init__()
        self.price_tolerance: Decimal = price_tolerance
        self.quantity_tolerance: Decimal = quantity_tolerance
        self.time_in_force_tolerance: timedelta = time_in_force_tolerance

    @staticmethod
    def zero_tolerance_order_reconciler() -> "ToleranceOrderReconciler":
        return ToleranceOrderReconciler(Decimal(0), Decimal(0), timedelta(seconds=0))

    def reconcile(
        self,
        _: ModelState,
        existing_orders: typing.Sequence[entropy.Order],
        desired_orders: typing.Sequence[entropy.Order],
    ) -> ReconciledOrders:
        remaining_existing_orders: typing.List[entropy.Order] = list(existing_orders)
        outcomes: ReconciledOrders = ReconciledOrders()
        for desired in desired_orders:
            acceptable = self.find_acceptable_order(desired, remaining_existing_orders)
            if acceptable is None:
                outcomes.to_place += [desired]
            else:
                outcomes.to_keep += [acceptable]
                outcomes.to_ignore += [desired]
                remaining_existing_orders.remove(acceptable)

        # By this point we have removed all acceptable existing orders, so those that remain
        # should be cancelled.
        outcomes.to_cancel = remaining_existing_orders

        in_count = len(existing_orders) + len(desired_orders)
        out_count = (
            len(outcomes.to_place)
            + len(outcomes.to_cancel)
            + len(outcomes.to_keep)
            + len(outcomes.to_ignore)
        )
        if in_count != out_count:
            raise Exception(
                f"Failure processing all desired orders. Count of orders in: {in_count}. Count of orders out: {out_count}."
            )

        return outcomes

    def find_acceptable_order(
        self, desired: entropy.Order, existing_orders: typing.Sequence[entropy.Order]
    ) -> typing.Optional[entropy.Order]:
        for existing in existing_orders:
            if self.is_within_tolderance(existing, desired):
                return existing
        return None

    def is_within_tolderance(
        self, existing: entropy.Order, desired: entropy.Order
    ) -> bool:
        if existing.side != desired.side:
            return False

        price_tolerance: Decimal = existing.price * self.price_tolerance
        if desired.price > (existing.price + price_tolerance):
            return False

        if desired.price < (existing.price - price_tolerance):
            return False

        quantity_tolerance: Decimal = existing.quantity * self.quantity_tolerance
        if desired.quantity > (existing.quantity + quantity_tolerance):
            return False

        if desired.quantity < (existing.quantity - quantity_tolerance):
            return False

        if desired.expiration > (existing.expiration + self.time_in_force_tolerance):
            return False

        if desired.expiration < (existing.expiration - self.time_in_force_tolerance):
            return False

        return True

    def __str__(self) -> str:
        return f"« ToleranceOrderReconciler [price tolerance: {self.price_tolerance}, quantity tolerance: {self.quantity_tolerance}, time-in-force tolerance: {self.time_in_force_tolerance}] »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .pairwiseelement import PairwiseElement
from ...modelstate import ModelState


# # 🥭 MinimumChargeElement class
#
# May modifiy an `Order`s price if it would result in too small a difference from the mid-price, meaning less
# of a charge if that `Order` is filled.
#
# Can take multiple minimum charges to work with pair-wise orders.
#
class MinimumChargeElement(PairwiseElement):
    def __init__(self, ratios: typing.Sequence[Decimal], from_bid_ask: bool) -> None:
        super().__init__()
        self.minimumcharge_ratios: typing.Sequence[Decimal] = ratios
        self.minimumcharge_from_bid_ask: bool = from_bid_ask

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--minimumcharge-ratio",
            type=Decimal,
            action="append",
            help="minimum fraction of the price to be accept as a spread",
        )
        parser.add_argument(
            "--minimumcharge-from-bid-ask",
            action="store_true",
            default=False,
            help="calculate minimum charge from bid or ask, not mid price (default: False, which will use the mid price)",
        )

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "MinimumChargeElement":
        minimumcharge_ratios: typing.Sequence[Decimal] = args.minimumcharge_ratio or [
            Decimal("0.0005")
        ]
        minimumcharge_from_bid_ask: bool = args.minimumcharge_from_bid_ask
        return MinimumChargeElement(minimumcharge_ratios, minimumcharge_from_bid_ask)

    def process_order_pair(
        self,
        context: entropy.Context,
        model_state: ModelState,
        index: int,
        buy: typing.Optional[entropy.Order],
        sell: typing.Optional[entropy.Order],
    ) -> typing.Tuple[typing.Optional[entropy.Order], typing.Optional[entropy.Order]]:
        # If no bias is explicitly specified for this element, just use the last specified bias.
        minimum_charge_ratio: Decimal = (
            self.minimumcharge_ratios[index]
            if index < len(self.minimumcharge_ratios)
            else self.minimumcharge_ratios[-1]
        )
        if minimum_charge_ratio == 0:
            # Zero minimum charge results in no changes to orders.
            return buy, sell

        # From Daffy on 26th July 2021: max(pyth_conf * 2, price * min_charge)
        # (Private chat link: https://discord.com/channels/@me/832570058861314048/869208592648134666)
        new_buy: typing.Optional[entropy.Order] = buy
        new_sell: typing.Optional[entropy.Order] = sell
        measurement_price: Decimal
        minimum_charge: Decimal
        current_charge: Decimal
        new_price: Decimal
        if buy is not None:
            measurement_price = (
                model_state.price.top_bid
                if self.minimumcharge_from_bid_ask
                else model_state.price.mid_price
            )
            minimum_charge = measurement_price * minimum_charge_ratio
            current_charge = measurement_price - buy.price
            if current_charge < minimum_charge:
                new_price = measurement_price - minimum_charge
                new_buy = buy.with_update(price=new_price)
                self._logger.debug(
                    f"""Order change - old BUY price {buy.price:,.8f} distance from {measurement_price:,.8f} would return {current_charge:,.8f} which is less than minimum charge {minimum_charge:,.8f}:
    Old: {buy}
    New: {new_buy}"""
                )

        if sell is not None:
            measurement_price = (
                model_state.price.top_ask
                if self.minimumcharge_from_bid_ask
                else model_state.price.mid_price
            )
            minimum_charge = measurement_price * minimum_charge_ratio
            current_charge = sell.price - measurement_price
            if current_charge < minimum_charge:
                new_price = measurement_price + minimum_charge
                new_sell = sell.with_update(price=new_price)
                self._logger.debug(
                    f"""Order change - old SELL price {sell.price:,.8f} distance from {measurement_price:,.8f} would return {current_charge:,.8f} which is less than minimum charge {minimum_charge:,.8f}:
    Old: {sell}
    New: {new_sell}"""
                )

        return new_buy, new_sell

    def __str__(self) -> str:
        return f"« MinimumChargeElement - minimum charge ratios: {self.minimumcharge_ratios} »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import logging
import entropy
import time
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from ..instrumentvalue import InstrumentValue
from ..modelstate import ModelState
from ..tokens import Token


# # 🥭 ModelStateBuilder class
#
# Base class for building a `ModelState` through polling or websockets.
#
class ModelStateBuilder(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @abc.abstractmethod
    def build(self, context: entropy.Context) -> ModelState:
        raise NotImplementedError(
            "ModelStateBuilder.build() is not implemented on the base type."
        )

    def __str__(self) -> str:
        return "« ModelStateBuilder »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 WebsocketModelStateBuilder class
#
# Base class for building a `ModelState` through polling.
#
class WebsocketModelStateBuilder(ModelStateBuilder):
    def __init__(self, model_state: ModelState) -> None:
        super().__init__()
        self.model_state: ModelState = model_state

    def build(self, context: entropy.Context) -> ModelState:
        return self.model_state

    def __str__(self) -> str:
        return f"« WebsocketModelStateBuilder for market '{self.model_state.market.fully_qualified_symbol}' »"


# # 🥭 PollingModelStateBuilder class
#
# Base class for building a `ModelState` through polling.
#
class PollingModelStateBuilder(ModelStateBuilder):
    def __init__(self) -> None:
        super().__init__()

    def build(self, context: entropy.Context) -> ModelState:
        started_at = time.time()
        built: ModelState = self.poll(context)
        time_taken = time.time() - started_at
        self._logger.debug(
            f"Poll for model state complete. Time taken: {time_taken:.2f} seconds."
        )
        return built

    @abc.abstractmethod
    def poll(self, context: entropy.Context) -> ModelState:
        raise NotImplementedError(
            "PollingModelStateBuilder.poll() is not implemented on the base type."
        )

    def from_values(
        self,
        order_owner: PublicKey,
        market: entropy.Market,
        group: entropy.Group,
        account: entropy.Account,
        price: entropy.Price,
        placed_orders_container: entropy.PlacedOrdersContainer,
        inventory: entropy.Inventory,
        orderbook: entropy.OrderBook,
        event_queue: entropy.EventQueue,
    ) -> ModelState:
        group_watcher: entropy.ManualUpdateWatcher[
            entropy.Group
        ] = entropy.ManualUpdateWatcher(group)
        account_watcher: entropy.ManualUpdateWatcher[
            entropy.Account
        ] = entropy.ManualUpdateWatcher(account)
        price_watcher: entropy.ManualUpdateWatcher[
            entropy.Price
        ] = entropy.ManualUpdateWatcher(price)
        placed_orders_container_watcher: entropy.ManualUpdateWatcher[
            entropy.PlacedOrdersContainer
        ] = entropy.ManualUpdateWatcher(placed_orders_container)
        inventory_watcher: entropy.ManualUpdateWatcher[
            entropy.Inventory
        ] = entropy.ManualUpdateWatcher(inventory)
        orderbook_watcher: entropy.ManualUpdateWatcher[
            entropy.OrderBook
        ] = entropy.ManualUpdateWatcher(orderbook)
        event_queue_watcher: entropy.ManualUpdateWatcher[
            entropy.EventQueue
        ] = entropy.ManualUpdateWatcher(event_queue)

        return ModelState(
            order_owner,
            market,
            group_watcher,
            account_watcher,
            price_watcher,
            placed_orders_container_watcher,
            inventory_watcher,
            orderbook_watcher,
            event_queue_watcher,
        )

    def __str__(self) -> str:
        return "« PollingModelStateBuilder »"


# # 🥭 SerumPollingModelStateBuilder class
#
# Polls Solana and builds a `ModelState` for a `SerumMarket`
#
class SerumPollingModelStateBuilder(PollingModelStateBuilder):
    def __init__(
        self,
        order_owner: PublicKey,
        market: entropy.SerumMarket,
        oracle: entropy.Oracle,
        group_address: PublicKey,
        cache_address: PublicKey,
        account_address: PublicKey,
        open_orders_address: PublicKey,
        base_inventory_token_account: entropy.TokenAccount,
        quote_inventory_token_account: entropy.TokenAccount,
    ) -> None:
        super().__init__()
        self.order_owner: PublicKey = order_owner
        self.market: entropy.SerumMarket = market
        self.oracle: entropy.Oracle = oracle

        self.group_address: PublicKey = group_address
        self.cache_address: PublicKey = cache_address
        self.account_address: PublicKey = account_address
        self.open_orders_address: PublicKey = open_orders_address
        self.base_inventory_token_account: entropy.TokenAccount = (
            base_inventory_token_account
        )
        self.quote_inventory_token_account: entropy.TokenAccount = (
            quote_inventory_token_account
        )

        # Serum always uses Tokens
        self.base_token: Token = Token.ensure(
            self.base_inventory_token_account.value.token
        )
        self.quote_token: Token = Token.ensure(
            self.quote_inventory_token_account.value.token
        )

    def poll(self, context: entropy.Context) -> ModelState:
        addresses: typing.List[PublicKey] = [
            self.group_address,
            self.cache_address,
            self.account_address,
            self.open_orders_address,
            self.base_inventory_token_account.address,
            self.quote_inventory_token_account.address,
            self.market.bids_address,
            self.market.asks_address,
            self.market.event_queue_address,
        ]
        account_infos: typing.Sequence[
            entropy.AccountInfo
        ] = entropy.AccountInfo.load_multiple(context, addresses)
        group: entropy.Group = entropy.Group.parse_with_context(
            context, account_infos[0]
        )
        cache: entropy.Cache = entropy.Cache.parse(account_infos[1])
        account: entropy.Account = entropy.Account.parse(account_infos[2], group, cache)
        placed_orders_container: entropy.PlacedOrdersContainer = (
            entropy.OpenOrders.parse(
                account_infos[3], self.market.base, self.market.quote
            )
        )

        # Serum markets don't accrue MNGO liquidity incentives
        mngo_accrued: InstrumentValue = InstrumentValue(
            group.liquidity_incentive_token, Decimal(0)
        )

        base_inventory_token_account = entropy.TokenAccount.parse(
            account_infos[4], self.base_token
        )
        quote_inventory_token_account = entropy.TokenAccount.parse(
            account_infos[5], self.quote_token
        )

        orderbook: entropy.OrderBook = self.market.parse_account_infos_to_orderbook(
            account_infos[6], account_infos[7]
        )

        event_queue: entropy.EventQueue = entropy.SerumEventQueue.parse(
            account_infos[8], self.base_token, self.quote_token
        )

        price: entropy.Price = self.oracle.fetch_price(context)

        available: Decimal = (
            base_inventory_token_account.value.value * price.mid_price
        ) + quote_inventory_token_account.value.value
        available_collateral: InstrumentValue = InstrumentValue(
            quote_inventory_token_account.value.token, available
        )
        inventory: entropy.Inventory = entropy.Inventory(
            entropy.InventorySource.SPL_TOKENS,
            mngo_accrued,
            available_collateral,
            base_inventory_token_account.value,
            quote_inventory_token_account.value,
        )

        return self.from_values(
            self.order_owner,
            self.market,
            group,
            account,
            price,
            placed_orders_container,
            inventory,
            orderbook,
            event_queue,
        )

    def __str__(self) -> str:
        return f"""« SerumPollingModelStateBuilder for market '{self.market.fully_qualified_symbol}' »"""


# # 🥭 SpotPollingModelStateBuilder class
#
# Polls Solana and builds a `ModelState` for a `SpotMarket`
#
class SpotPollingModelStateBuilder(PollingModelStateBuilder):
    def __init__(
        self,
        order_owner: PublicKey,
        market: entropy.SpotMarket,
        oracle: entropy.Oracle,
        group_address: PublicKey,
        cache_address: PublicKey,
        account_address: PublicKey,
        open_orders_address: PublicKey,
        all_open_orders_addresses: typing.Sequence[PublicKey],
    ) -> None:
        super().__init__()
        self.order_owner: PublicKey = order_owner
        self.market: entropy.SpotMarket = market
        self.oracle: entropy.Oracle = oracle

        self.group_address: PublicKey = group_address
        self.cache_address: PublicKey = cache_address
        self.account_address: PublicKey = account_address
        self.open_orders_address: PublicKey = open_orders_address
        self.all_open_orders_addresses: typing.Sequence[
            PublicKey
        ] = all_open_orders_addresses

    def poll(self, context: entropy.Context) -> ModelState:
        addresses: typing.List[PublicKey] = [
            self.group_address,
            self.cache_address,
            self.account_address,
            self.market.bids_address,
            self.market.asks_address,
            self.market.event_queue_address,
            *self.all_open_orders_addresses,
        ]
        account_infos: typing.Sequence[
            entropy.AccountInfo
        ] = entropy.AccountInfo.load_multiple(context, addresses)
        group: entropy.Group = entropy.Group.parse_with_context(
            context, account_infos[0]
        )
        cache: entropy.Cache = entropy.Cache.parse(account_infos[1])
        account: entropy.Account = entropy.Account.parse(account_infos[2], group, cache)

        # Update our stash of OpenOrders addresses for next time, in case new OpenOrders accounts were added
        self.all_open_orders_addresses = account.spot_open_orders

        spot_open_orders_account_infos_by_address = {
            str(account_info.address): account_info
            for account_info in account_infos[6:]
        }

        all_open_orders: typing.Dict[str, entropy.OpenOrders] = {}
        for basket_token in account.slots:
            if (
                basket_token.spot_open_orders is not None
                and str(basket_token.spot_open_orders)
                in spot_open_orders_account_infos_by_address
            ):
                account_info: entropy.AccountInfo = (
                    spot_open_orders_account_infos_by_address[
                        str(basket_token.spot_open_orders)
                    ]
                )
                open_orders: entropy.OpenOrders = entropy.OpenOrders.parse(
                    account_info,
                    Token.ensure(basket_token.base_instrument),
                    account.shared_quote_token,
                )
                all_open_orders[str(basket_token.spot_open_orders)] = open_orders

        placed_orders_container: entropy.PlacedOrdersContainer = all_open_orders[
            str(self.open_orders_address)
        ]

        # Spot markets don't accrue MNGO liquidity incentives
        mngo_accrued: InstrumentValue = InstrumentValue(
            group.liquidity_incentive_token, Decimal(0)
        )

        base_value = entropy.InstrumentValue.find_by_symbol(
            account.net_values, self.market.base.symbol
        )
        quote_value = entropy.InstrumentValue.find_by_symbol(
            account.net_values, self.market.quote.symbol
        )

        frame = account.to_dataframe(group, all_open_orders, cache)
        available_collateral: InstrumentValue = account.init_health(frame)
        inventory: entropy.Inventory = entropy.Inventory(
            entropy.InventorySource.ACCOUNT,
            mngo_accrued,
            available_collateral,
            base_value,
            quote_value,
        )

        orderbook: entropy.OrderBook = self.market.parse_account_infos_to_orderbook(
            account_infos[3], account_infos[4]
        )

        event_queue: entropy.EventQueue = entropy.SerumEventQueue.parse(
            account_infos[5], self.market.base, self.market.quote
        )

        price: entropy.Price = self.oracle.fetch_price(context)

        return self.from_values(
            self.order_owner,
            self.market,
            group,
            account,
            price,
            placed_orders_container,
            inventory,
            orderbook,
            event_queue,
        )

    def __str__(self) -> str:
        return f"""« SpotPollingModelStateBuilder for market '{self.market.fully_qualified_symbol}' »"""


# # 🥭 PerpPollingModelStateBuilder class
#
# Polls Solana and builds a `ModelState` for a `PerpMarket`
#
class PerpPollingModelStateBuilder(PollingModelStateBuilder):
    def __init__(
        self,
        order_owner: PublicKey,
        market: entropy.PerpMarket,
        oracle: entropy.Oracle,
        group_address: PublicKey,
        cache_address: PublicKey,
        all_open_orders_addresses: typing.Sequence[PublicKey],
    ) -> None:
        super().__init__()
        self.order_owner: PublicKey = order_owner
        self.market: entropy.PerpMarket = market
        self.oracle: entropy.Oracle = oracle

        self.group_address: PublicKey = group_address
        self.cache_address: PublicKey = cache_address

        self.all_open_orders_addresses: typing.Sequence[
            PublicKey
        ] = all_open_orders_addresses

    def poll(self, context: entropy.Context) -> ModelState:
        addresses: typing.List[PublicKey] = [
            self.group_address,
            self.cache_address,
            self.order_owner,
            self.market.underlying_perp_market.bids,
            self.market.underlying_perp_market.asks,
            self.market.event_queue_address,
            *self.all_open_orders_addresses,
        ]
        account_infos: typing.Sequence[
            entropy.AccountInfo
        ] = entropy.AccountInfo.load_multiple(context, addresses)
        group: entropy.Group = entropy.Group.parse_with_context(
            context, account_infos[0]
        )
        cache: entropy.Cache = entropy.Cache.parse(account_infos[1])
        account: entropy.Account = entropy.Account.parse(account_infos[2], group, cache)

        # Update our stash of OpenOrders addresses for next time, in case new OpenOrders accounts were added
        self.all_open_orders_addresses = account.spot_open_orders

        spot_open_orders_account_infos_by_address = {
            str(account_info.address): account_info
            for account_info in account_infos[6:]
        }

        all_open_orders: typing.Dict[str, entropy.OpenOrders] = {}
        for basket_token in account.slots:
            if (
                basket_token.spot_open_orders is not None
                and str(basket_token.spot_open_orders)
                in spot_open_orders_account_infos_by_address
            ):
                account_info: entropy.AccountInfo = (
                    spot_open_orders_account_infos_by_address[
                        str(basket_token.spot_open_orders)
                    ]
                )
                open_orders: entropy.OpenOrders = entropy.OpenOrders.parse(
                    account_info,
                    Token.ensure(basket_token.base_instrument),
                    account.shared_quote_token,
                )
                all_open_orders[str(basket_token.spot_open_orders)] = open_orders

        slot = group.slot_by_perp_market_address(self.market.address)
        perp_account = account.perp_accounts_by_index[slot.index]
        if perp_account is None:
            raise Exception(
                f"Could not find perp account at index {slot.index} of account {account.address}."
            )
        placed_orders_container: entropy.PlacedOrdersContainer = (
            perp_account.open_orders
        )

        base_lots = perp_account.base_position
        base_value = self.market.lot_size_converter.base_size_lots_to_number(base_lots)
        base_token_value = entropy.InstrumentValue(self.market.base, base_value)
        quote_token_value = account.shared_quote.net_value
        frame = account.to_dataframe(group, all_open_orders, cache)
        available_collateral: InstrumentValue = account.init_health(frame)
        inventory: entropy.Inventory = entropy.Inventory(
            entropy.InventorySource.ACCOUNT,
            perp_account.mngo_accrued,
            available_collateral,
            base_token_value,
            quote_token_value,
        )

        orderbook: entropy.OrderBook = self.market.parse_account_infos_to_orderbook(
            account_infos[3], account_infos[4]
        )

        event_queue: entropy.EventQueue = entropy.PerpEventQueue.parse(
            account_infos[5], self.market.lot_size_converter
        )

        price: entropy.Price = self.oracle.fetch_price(context)

        return self.from_values(
            self.order_owner,
            self.market,
            group,
            account,
            price,
            placed_orders_container,
            inventory,
            orderbook,
            event_queue,
        )

    def __str__(self) -> str:
        return f"""« PerpPollingModelStateBuilder for market '{self.market.fully_qualified_symbol}' »"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .pairwiseelement import PairwiseElement
from ...modelstate import ModelState


# # 🥭 BiasQuoteOnPositionElement class
#
# Modifies an `Order`s price based on current inventory. Uses a `bias` to shift the price to sell
# more (if too much inventory) or buy more (if too little inventory).
#
# Can take multiple bias factors to work with pair-wise orders.
#
class BiasQuoteOnPositionElement(PairwiseElement):
    def __init__(self, biases: typing.Sequence[Decimal]) -> None:
        super().__init__()
        self.biases: typing.Sequence[Decimal] = biases

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--biasquoteonposition-bias",
            type=Decimal,
            action="append",
            help="bias to apply to quotes based on inventory position",
        )

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "BiasQuoteOnPositionElement":
        biases: typing.Sequence[Decimal] = args.biasquoteonposition_bias or [Decimal(0)]
        return BiasQuoteOnPositionElement(biases)

    def process_order_pair(
        self,
        context: entropy.Context,
        model_state: ModelState,
        index: int,
        buy: typing.Optional[entropy.Order],
        sell: typing.Optional[entropy.Order],
    ) -> typing.Tuple[typing.Optional[entropy.Order], typing.Optional[entropy.Order]]:
        # If no bias is explicitly specified for this element, just use the last specified bias.
        bias: Decimal = (
            self.biases[index] if index < len(self.biases) else self.biases[-1]
        )
        if bias == 0:
            # Zero bias results in no changes to orders.
            return buy, sell

        new_buy: typing.Optional[entropy.Order] = None
        new_sell: typing.Optional[entropy.Order] = None
        if buy is not None:
            new_buy = self.bias_order(buy, bias, model_state.inventory.base.value)

        if sell is not None:
            new_sell = self.bias_order(sell, bias, model_state.inventory.base.value)

        return new_buy, new_sell

    # From Daffy on 20th August 2021:
    #  Formula to adjust price might look like this `pyth_price * (1 + (curr_pos / size) * pos_lean)`
    #  where pos_lean is a negative number
    #
    #  size is the standard size you're quoting which I believe comes from the position-size-ratio
    #
    #  So if my standard size I'm quoting is 0.0002 BTC, my current position is +0.0010 BTC, and pos_lean
    #  is -0.0001, you would move your quotes down by 0.0005 (or 5bps)
    # (Private chat link: https://discord.com/channels/@me/832570058861314048/878343278523723787)
    def bias_order(
        self,
        order: entropy.Order,
        inventory_bias: Decimal,
        base_inventory_value: Decimal,
    ) -> entropy.Order:
        bias_factor = inventory_bias * -1
        bias = 1 + ((base_inventory_value / order.quantity) * bias_factor)
        new_price: Decimal = order.price * bias
        new_order: entropy.Order = order.with_update(price=new_price)
        bias_description = "BUY more" if bias > 1 else "SELL more"
        self._logger.debug(
            f"""Order change - bias {inventory_bias} on inventory {base_inventory_value} / {order.quantity} creates a ({bias_description}) bias factor of {bias}:
    Old: {order}
    New: {new_order}"""
        )
        return new_order

    def __str__(self) -> str:
        return f"« BiasQuoteOnPositionElement - biases: {self.biases} »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import enum


# # 🥭 Version enum
#
# This is used to keep track of which version of the layout struct was used to load the
# data.
#


class Version(enum.Enum):
    UNSPECIFIED = 0
    V1 = 1
    V2 = 2
    V3 = 3
    V4 = 4
    V5 = 5



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import rx.operators
import typing

from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .context import Context
from .lotsizeconverter import LotSizeConverter
from .markets import InventorySource, MarketType, Market
from .observables import Disposable
from .orders import Order, OrderBook
from .tokens import Instrument, Token
from .websocketsubscription import (
    SharedWebSocketSubscriptionManager,
    WebSocketAccountSubscription,
)


class Event(typing.Protocol):
    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        raise NotImplementedError(
            "Event.accounts_to_crank is not implemented on the Protocol."
        )


class FillEvent(Event, typing.Protocol):
    pass


# # 🥭 LoadedMarket class
#
# This class describes a crypto market. It *must* have an address, a base token and a quote token.
#
class LoadedMarket(Market):
    def __init__(
        self,
        type: MarketType,
        program_address: PublicKey,
        address: PublicKey,
        inventory_source: InventorySource,
        base: Instrument,
        quote: Token,
        lot_size_converter: LotSizeConverter,
    ) -> None:
        super().__init__(
            type,
            program_address,
            address,
            inventory_source,
            base,
            quote,
            lot_size_converter,
        )

    @property
    @abc.abstractproperty
    def bids_address(self) -> PublicKey:
        raise NotImplementedError(
            "LoadedMarket.bids_address() is not implemented on the base type."
        )

    @property
    @abc.abstractproperty
    def asks_address(self) -> PublicKey:
        raise NotImplementedError(
            "LoadedMarket.asks_address() is not implemented on the base type."
        )

    @property
    @abc.abstractproperty
    def event_queue_address(self) -> PublicKey:
        raise NotImplementedError(
            "LoadedMarket.event_queue_address() is not implemented on the base type."
        )

    @abc.abstractmethod
    def on_fill(
        self, context: Context, handler: typing.Callable[[FillEvent], None]
    ) -> Disposable:
        raise NotImplementedError(
            "LoadedMarket.on_fill() is not implemented on the base type."
        )

    @abc.abstractmethod
    def on_event(
        self, context: Context, handler: typing.Callable[[Event], None]
    ) -> Disposable:
        raise NotImplementedError(
            "LoadedMarket.on_event() is not implemented on the base type."
        )

    def on_orderbook_change(
        self, context: Context, handler: typing.Callable[[OrderBook], None]
    ) -> Disposable:
        disposer = Disposable()

        [bids_info, asks_info] = AccountInfo.load_multiple(
            context, [self.bids_address, self.asks_address]
        )

        stored: OrderBook = self.parse_account_infos_to_orderbook(bids_info, asks_info)

        def _update_bids(account_info: AccountInfo) -> OrderBook:
            new_bids = self.parse_account_info_to_orders(account_info)
            stored.bids = new_bids
            return OrderBook(
                self.symbol, self.lot_size_converter, new_bids, stored.asks
            )

        def _update_asks(account_info: AccountInfo) -> OrderBook:
            new_asks = self.parse_account_info_to_orders(account_info)
            stored.asks = new_asks
            return OrderBook(
                self.symbol, self.lot_size_converter, stored.bids, new_asks
            )

        websocket = SharedWebSocketSubscriptionManager(context)
        disposer.add_disposable(websocket)

        bids_subscription = WebSocketAccountSubscription[OrderBook](
            context, self.bids_address, _update_bids
        )
        websocket.add(bids_subscription)

        asks_subscription = WebSocketAccountSubscription[OrderBook](
            context, self.asks_address, _update_asks
        )
        websocket.add(asks_subscription)

        orderbook_changes = bids_subscription.publisher.pipe(
            rx.operators.merge(asks_subscription.publisher)
        )

        individual_event_subscription = orderbook_changes.subscribe(on_next=handler)
        disposer.add_disposable(individual_event_subscription)

        websocket.open()

        return disposer

    @abc.abstractmethod
    def parse_account_info_to_orders(
        self, account_info: AccountInfo
    ) -> typing.Sequence[Order]:
        raise NotImplementedError(
            "LoadedMarket.parse_account_info_to_orders() is not implemented on the base type."
        )

    def parse_account_infos_to_orderbook(
        self, bids_account_info: AccountInfo, asks_account_info: AccountInfo
    ) -> OrderBook:
        bids_orderbook = self.parse_account_info_to_orders(bids_account_info)
        asks_orderbook = self.parse_account_info_to_orders(asks_account_info)
        return OrderBook(
            self.symbol, self.lot_size_converter, bids_orderbook, asks_orderbook
        )

    def fetch_orderbook(self, context: Context) -> OrderBook:
        [bids_info, asks_info] = AccountInfo.load_multiple(
            context, [self.bids_address, self.asks_address]
        )
        return self.parse_account_infos_to_orderbook(bids_info, asks_info)
